"""
Module for managing the Salt schedule on a minion

Requires that python-dateutil is installed on the minion.

.. versionadded:: 2014.7.0

"""

import copy as pycopy
import datetime
import logging
import os

import yaml

import salt.utils.event
import salt.utils.files
import salt.utils.odict
import salt.utils.yaml

try:
    import dateutil.parser as dateutil_parser

    _WHEN_SUPPORTED = True
    _RANGE_SUPPORTED = True
except ImportError:
    _WHEN_SUPPORTED = False
    _RANGE_SUPPORTED = False

__proxyenabled__ = ["*"]

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list", "reload_": "reload"}

SCHEDULE_CONF = [
    "name",
    "maxrunning",
    "function",
    "splay",
    "range",
    "when",
    "once",
    "once_fmt",
    "returner",
    "jid_include",
    "args",
    "kwargs",
    "_seconds",
    "seconds",
    "minutes",
    "hours",
    "days",
    "enabled",
    "return_job",
    "metadata",
    "cron",
    "until",
    "after",
    "return_config",
    "return_kwargs",
    "run_on_start",
    "skip_during_range",
    "run_after_skip_range",
]


def _get_schedule_config_file():
    """
    Return the minion schedule configuration file
    """
    config_dir = __opts__.get("conf_dir", None)
    if config_dir is None and "conf_file" in __opts__:
        config_dir = os.path.dirname(__opts__["conf_file"])
    if config_dir is None:
        config_dir = salt.syspaths.CONFIG_DIR

    minion_d_dir = os.path.join(
        config_dir,
        os.path.dirname(
            __opts__.get(
                "default_include",
                salt.config.DEFAULT_MINION_OPTS["default_include"],
            )
        ),
    )

    if not os.path.isdir(config_dir):
        os.makedirs(config_dir)

    if not os.path.isdir(minion_d_dir):
        os.makedirs(minion_d_dir)

    return os.path.join(minion_d_dir, "_schedule.conf")


def list_(
    show_all=False, show_disabled=True, where=None, return_yaml=True, offline=False
):
    """
    List the jobs currently scheduled on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.list

        # Show all jobs including hidden internal jobs
        salt '*' schedule.list show_all=True

        # Hide disabled jobs from list of jobs
        salt '*' schedule.list show_disabled=False

    """

    def _get_saved():
        schedule = {}
        schedule_config = _get_schedule_config_file()
        if os.path.exists(schedule_config):
            with salt.utils.files.fopen(schedule_config) as fp_:
                schedule_yaml = fp_.read()
                if schedule_yaml:
                    schedule_contents = yaml.safe_load(schedule_yaml)
                    schedule = schedule_contents.get("schedule", {})
        return schedule

    schedule = {}
    if offline:
        schedule = _get_saved()
        saved_schedule = pycopy.deepcopy(schedule)
    else:
        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](
                    {"func": "list", "where": where}, "manage_schedule"
                )
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_list_complete", wait=30
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret = {}
            ret["comment"] = "Event module not available. Schedule list failed."
            ret["result"] = True
            log.debug("Event module not available. Schedule list failed.")
            return ret

        saved_schedule = _get_saved()

    _hidden = ["enabled", "skip_function", "skip_during_range"]
    for job in list(schedule.keys()):  # iterate over a copy since we will mutate it
        if job in _hidden:
            continue

        # Default jobs added by salt begin with __
        # by default hide them unless show_all is True.
        if job.startswith("__") and not show_all:
            del schedule[job]
            continue

        # if enabled is not included in the job,
        # assume job is enabled.
        if "enabled" not in schedule[job]:
            schedule[job]["enabled"] = True

        for item in pycopy.copy(schedule[job]):
            if item not in SCHEDULE_CONF:
                del schedule[job][item]
                continue
            if schedule[job][item] is None:
                del schedule[job][item]
                continue
            if schedule[job][item] == "true":
                schedule[job][item] = True
            if schedule[job][item] == "false":
                schedule[job][item] = False

        # if the job is disabled and show_disabled is False, skip job
        if not show_disabled and not schedule[job]["enabled"]:
            del schedule[job]
            continue

        if "_seconds" in schedule[job]:
            # remove _seconds from the listing
            del schedule[job]["_seconds"]

    if return_yaml:
        # Indicate whether the scheduled job is saved
        # to the minion configuration.
        for item in schedule:
            if isinstance(schedule[item], dict):
                if item in saved_schedule:
                    schedule[item]["saved"] = True
                else:
                    schedule[item]["saved"] = False
        tmp = {"schedule": schedule}
        return salt.utils.yaml.safe_dump(tmp, default_flow_style=False)
    else:
        return schedule


def is_enabled(name=None):
    """
    List a Job only if its enabled

    If job is not specified, indicate
    if the scheduler is enabled or disabled.

    .. versionadded:: 2015.5.3

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.is_enabled name=job_name
        salt '*' schedule.is_enabled
    """

    current_schedule = __salt__["schedule.list"](show_all=False, return_yaml=False)
    if not name:
        return current_schedule.get("enabled", True)
    else:
        if name in current_schedule:
            return current_schedule[name]
        else:
            return {}


def purge(**kwargs):
    """
    Purge all the jobs currently scheduled on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.purge

        # Purge jobs on Salt minion
        salt '*' schedule.purge

    """

    ret = {"comment": [], "changes": {}, "result": True}

    current_schedule = list_(
        show_all=True, return_yaml=False, offline=kwargs.get("offline")
    )
    for name in pycopy.deepcopy(current_schedule):
        if name == "enabled":
            continue
        if name.startswith("__"):
            continue

        if "test" in kwargs and kwargs["test"]:
            ret["result"] = True
            ret["comment"].append(f"Job: {name} would be deleted from schedule.")
        else:
            if kwargs.get("offline"):
                del current_schedule[name]

                ret["comment"].append(f"Deleted job: {name} from schedule.")
                ret["changes"][name] = "removed"

            else:
                persist = kwargs.get("persist", True)
                try:
                    with salt.utils.event.get_event(
                        "minion", opts=__opts__
                    ) as event_bus:
                        res = __salt__["event.fire"](
                            {"name": name, "func": "delete", "persist": persist},
                            "manage_schedule",
                        )
                        if res:
                            event_ret = event_bus.get_event(
                                tag="/salt/minion/minion_schedule_delete_complete",
                                wait=30,
                            )
                            if event_ret and event_ret["complete"]:
                                _schedule_ret = event_ret["schedule"]
                                if name not in _schedule_ret:
                                    ret["result"] = True
                                    ret["changes"][name] = "removed"
                                    ret["comment"].append(
                                        f"Deleted job: {name} from schedule."
                                    )
                                else:
                                    ret["comment"].append(
                                        "Failed to delete job {} from schedule.".format(
                                            name
                                        )
                                    )
                                    ret["result"] = True

                except KeyError:
                    # Effectively a no-op, since we can't really return without an event system
                    ret["comment"] = "Event module not available. Schedule add failed."
                    ret["result"] = True

    # wait until the end to write file in offline mode
    if kwargs.get("offline"):
        schedule_conf = _get_schedule_config_file()

        try:
            with salt.utils.files.fopen(schedule_conf, "wb+") as fp_:
                fp_.write(
                    salt.utils.stringutils.to_bytes(
                        salt.utils.yaml.safe_dump({"schedule": current_schedule})
                    )
                )
        except OSError:
            log.error(
                "Failed to persist the updated schedule",
                exc_info_on_loglevel=logging.DEBUG,
            )

    return ret


def delete(name, **kwargs):
    """
    Delete a job from the minion's schedule

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.delete job1

        # Delete job on Salt minion when the Salt minion is not running
        salt '*' schedule.delete job1

    """

    ret = {
        "comment": f"Failed to delete job {name} from schedule.",
        "result": False,
        "changes": {},
    }

    if not name:
        ret["comment"] = "Job name is required."

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be deleted from schedule."
        ret["result"] = True
    else:
        if kwargs.get("offline"):
            current_schedule = list_(
                show_all=True,
                where="opts",
                return_yaml=False,
                offline=kwargs.get("offline"),
            )

            del current_schedule[name]

            schedule_conf = _get_schedule_config_file()

            try:
                with salt.utils.files.fopen(schedule_conf, "wb+") as fp_:
                    fp_.write(
                        salt.utils.stringutils.to_bytes(
                            salt.utils.yaml.safe_dump({"schedule": current_schedule})
                        )
                    )
            except OSError:
                log.error(
                    "Failed to persist the updated schedule",
                    exc_info_on_loglevel=logging.DEBUG,
                )

            ret["result"] = True
            ret["comment"] = f"Deleted Job {name} from schedule."
            ret["changes"][name] = "removed"
        else:
            persist = kwargs.get("persist", True)

            if name in list_(
                show_all=True,
                where="opts",
                return_yaml=False,
                offline=kwargs.get("offline"),
            ):
                event_data = {"name": name, "func": "delete", "persist": persist}
            elif name in list_(
                show_all=True,
                where="pillar",
                return_yaml=False,
                offline=kwargs.get("offline"),
            ):
                event_data = {
                    "name": name,
                    "where": "pillar",
                    "func": "delete",
                    "persist": False,
                }
            else:
                ret["comment"] = f"Job {name} does not exist."
                return ret

            try:
                with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                    res = __salt__["event.fire"](event_data, "manage_schedule")
                    if res:
                        event_ret = event_bus.get_event(
                            tag="/salt/minion/minion_schedule_delete_complete",
                            wait=30,
                        )
                        if event_ret and event_ret["complete"]:
                            schedule = event_ret["schedule"]
                            if name not in schedule:
                                ret["result"] = True
                                ret["comment"] = "Deleted Job {} from schedule.".format(
                                    name
                                )
                                ret["changes"][name] = "removed"
                            else:
                                ret["comment"] = (
                                    "Failed to delete job {} from schedule.".format(
                                        name
                                    )
                                )
                            return ret
            except KeyError:
                # Effectively a no-op, since we can't really return without an event system
                ret["comment"] = "Event module not available. Schedule add failed."
    return ret


def build_schedule_item(name, **kwargs):
    """
    Build a schedule job

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.build_schedule_item job1 function='test.ping' seconds=3600
    """

    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False
        return ret

    schedule = {}
    schedule[name] = salt.utils.odict.OrderedDict()
    schedule[name]["function"] = kwargs["function"]

    time_conflict = False
    for item in ["seconds", "minutes", "hours", "days"]:
        if item in kwargs and "when" in kwargs:
            time_conflict = True

        if item in kwargs and "cron" in kwargs:
            time_conflict = True

    if time_conflict:
        ret["result"] = False
        ret["comment"] = (
            'Unable to use "seconds", "minutes", "hours", or "days" with "when" or'
            ' "cron" options.'
        )
        return ret

    if "when" in kwargs and "cron" in kwargs:
        ret["result"] = False
        ret["comment"] = 'Unable to use "when" and "cron" options together.  Ignoring.'
        return ret

    for item in ["seconds", "minutes", "hours", "days"]:
        if item in kwargs:
            schedule[name][item] = kwargs[item]

    if "return_job" in kwargs:
        schedule[name]["return_job"] = kwargs["return_job"]

    if "metadata" in kwargs:
        schedule[name]["metadata"] = kwargs["metadata"]

    if "job_args" in kwargs:
        if isinstance(kwargs["job_args"], list):
            schedule[name]["args"] = kwargs["job_args"]
        else:
            ret["result"] = False
            ret["comment"] = "job_args is not a list. please correct and try again."
            return ret

    if "job_kwargs" in kwargs:
        if isinstance(kwargs["job_kwargs"], dict):
            schedule[name]["kwargs"] = kwargs["job_kwargs"]
        else:
            ret["result"] = False
            ret["comment"] = "job_kwargs is not a dict. please correct and try again."
            return ret

    if "maxrunning" in kwargs:
        schedule[name]["maxrunning"] = kwargs["maxrunning"]
    else:
        schedule[name]["maxrunning"] = 1

    if "name" in kwargs:
        schedule[name]["name"] = kwargs["name"]
    else:
        schedule[name]["name"] = name

    if "enabled" in kwargs:
        schedule[name]["enabled"] = kwargs["enabled"]
    else:
        schedule[name]["enabled"] = True

    schedule[name]["jid_include"] = kwargs.get("jid_include", True)

    if "splay" in kwargs:
        if isinstance(kwargs["splay"], dict):
            # Ensure ordering of start and end arguments
            schedule[name]["splay"] = salt.utils.odict.OrderedDict()
            schedule[name]["splay"]["start"] = kwargs["splay"]["start"]
            schedule[name]["splay"]["end"] = kwargs["splay"]["end"]
        else:
            schedule[name]["splay"] = kwargs["splay"]

    if "when" in kwargs:
        if not _WHEN_SUPPORTED:
            ret["result"] = False
            ret["comment"] = 'Missing dateutil.parser, "when" is unavailable.'
            return ret
        else:
            validate_when = kwargs["when"]
            if not isinstance(validate_when, list):
                validate_when = [validate_when]
            for _when in validate_when:
                try:
                    dateutil_parser.parse(_when)
                except ValueError:
                    ret["result"] = False
                    ret["comment"] = 'Schedule item {} for "when" in invalid.'.format(
                        _when
                    )
                    return ret

    for item in [
        "range",
        "when",
        "once",
        "once_fmt",
        "cron",
        "returner",
        "after",
        "return_config",
        "return_kwargs",
        "until",
        "run_on_start",
        "skip_during_range",
    ]:
        if item in kwargs:
            schedule[name][item] = kwargs[item]

    return schedule[name]


def add(name, **kwargs):
    """
    Add a job to the schedule

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.add job1 function='test.ping' seconds=3600
        # If function have some arguments, use job_args
        salt '*' schedule.add job2 function='cmd.run' job_args="['date >> /tmp/date.log']" seconds=60

        # Add job to Salt minion when the Salt minion is not running
        salt '*' schedule.add job1 function='test.ping' seconds=3600 offline=True

    """

    ret = {
        "comment": f"Failed to add job {name} to schedule.",
        "result": False,
        "changes": {},
    }
    current_schedule = list_(
        show_all=True, return_yaml=False, offline=kwargs.get("offline")
    )

    if name in current_schedule:
        ret["comment"] = f"Job {name} already exists in schedule."
        ret["result"] = False
        return ret

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    time_conflict = False
    for item in ["seconds", "minutes", "hours", "days"]:
        if item in kwargs and "when" in kwargs:
            time_conflict = True
        if item in kwargs and "cron" in kwargs:
            time_conflict = True

    if time_conflict:
        ret["comment"] = (
            'Error: Unable to use "seconds", "minutes", "hours", or "days" with "when"'
            ' or "cron" options.'
        )
        return ret

    if "when" in kwargs and "cron" in kwargs:
        ret["comment"] = 'Unable to use "when" and "cron" options together.  Ignoring.'
        return ret

    persist = kwargs.get("persist", True)

    _new = build_schedule_item(name, **kwargs)
    if "result" in _new and not _new["result"]:
        return _new

    schedule_data = {}
    schedule_data[name] = _new

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be added to schedule."
        ret["result"] = True
    else:
        if kwargs.get("offline"):
            current_schedule.update(schedule_data)

            schedule_conf = _get_schedule_config_file()

            try:
                with salt.utils.files.fopen(schedule_conf, "wb+") as fp_:
                    fp_.write(
                        salt.utils.stringutils.to_bytes(
                            salt.utils.yaml.safe_dump({"schedule": current_schedule})
                        )
                    )
            except OSError:
                log.error(
                    "Failed to persist the updated schedule",
                    exc_info_on_loglevel=logging.DEBUG,
                )

            ret["result"] = True
            ret["comment"] = f"Added job: {name} to schedule."
            ret["changes"][name] = "added"
        else:
            try:
                with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                    res = __salt__["event.fire"](
                        {
                            "name": name,
                            "schedule": schedule_data,
                            "func": "add",
                            "persist": persist,
                        },
                        "manage_schedule",
                    )
                    if res:
                        event_ret = event_bus.get_event(
                            tag="/salt/minion/minion_schedule_add_complete",
                            wait=30,
                        )
                        if event_ret and event_ret["complete"]:
                            schedule = event_ret["schedule"]
                            if name in schedule:
                                ret["result"] = True
                                ret["comment"] = "Added job: {} to schedule.".format(
                                    name
                                )
                                ret["changes"][name] = "added"
                                return ret
            except KeyError:
                # Effectively a no-op, since we can't really return without an event system
                ret["comment"] = "Event module not available. Schedule add failed."
    return ret


def modify(name, **kwargs):
    """
    Modify an existing job in the schedule

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.modify job1 function='test.ping' seconds=3600

        # Modify job on Salt minion when the Salt minion is not running
        salt '*' schedule.modify job1 function='test.ping' seconds=3600 offline=True

    """

    ret = {"comment": "", "changes": {}, "result": True}

    time_conflict = False
    for item in ["seconds", "minutes", "hours", "days"]:
        if item in kwargs and "when" in kwargs:
            time_conflict = True

        if item in kwargs and "cron" in kwargs:
            time_conflict = True

    if time_conflict:
        ret["result"] = False
        ret["comment"] = (
            'Error: Unable to use "seconds", "minutes", "hours", or "days" with "when"'
            " option."
        )
        return ret

    if "when" in kwargs and "cron" in kwargs:
        ret["result"] = False
        ret["comment"] = 'Unable to use "when" and "cron" options together.  Ignoring.'
        return ret

    current_schedule = list_(
        show_all=True, return_yaml=False, offline=kwargs.get("offline")
    )

    if name not in current_schedule:
        ret["comment"] = f"Job {name} does not exist in schedule."
        ret["result"] = False
        return ret

    _current = current_schedule[name]

    if "function" not in kwargs:
        kwargs["function"] = _current.get("function")

    # Remove the auto generated _seconds value
    if "_seconds" in _current:
        _current["seconds"] = _current.pop("_seconds")

    # Copy _current _new, then update values from kwargs
    _new = build_schedule_item(name, **kwargs)

    # Remove test from kwargs, it's not a valid schedule option
    _new.pop("test", None)

    if "result" in _new and not _new["result"]:
        return _new

    if _new == _current:
        ret["comment"] = f"Job {name} in correct state"
        return ret

    ret["changes"][name] = {
        "old": salt.utils.odict.OrderedDict(_current),
        "new": salt.utils.odict.OrderedDict(_new),
    }

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be modified in schedule."
    else:
        if kwargs.get("offline"):
            current_schedule[name].update(_new)

            schedule_conf = _get_schedule_config_file()

            try:
                with salt.utils.files.fopen(schedule_conf, "wb+") as fp_:
                    fp_.write(
                        salt.utils.stringutils.to_bytes(
                            salt.utils.yaml.safe_dump({"schedule": current_schedule})
                        )
                    )
            except OSError:
                log.error(
                    "Failed to persist the updated schedule",
                    exc_info_on_loglevel=logging.DEBUG,
                )

            ret["result"] = True
            ret["comment"] = f"Modified job: {name} in schedule."

        else:
            persist = kwargs.get("persist", True)
            if name in list_(show_all=True, where="opts", return_yaml=False):
                event_data = {
                    "name": name,
                    "schedule": _new,
                    "func": "modify",
                    "persist": persist,
                }
            elif name in list_(show_all=True, where="pillar", return_yaml=False):
                event_data = {
                    "name": name,
                    "schedule": _new,
                    "where": "pillar",
                    "func": "modify",
                    "persist": False,
                }

            out = __salt__["event.fire"](event_data, "manage_schedule")
            if out:
                ret["comment"] = f"Modified job: {name} in schedule."
            else:
                ret["comment"] = f"Failed to modify job {name} in schedule."
                ret["result"] = False
    return ret


def run_job(name, force=False):
    """
    Run a scheduled job on the minion immediately

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.run_job job1

        salt '*' schedule.run_job job1 force=True
        Force the job to run even if it is disabled.
    """

    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    schedule = list_(show_all=True, return_yaml=False)
    if name in schedule:
        data = schedule[name]
        if "enabled" in data and not data["enabled"] and not force:
            ret["comment"] = f"Job {name} is disabled."
        else:
            out = __salt__["event.fire"](
                {"name": name, "func": "run_job"}, "manage_schedule"
            )
            if out:
                ret["comment"] = f"Scheduling Job {name} on minion."
            else:
                ret["comment"] = f"Failed to run job {name} on minion."
                ret["result"] = False
    else:
        ret["comment"] = f"Job {name} does not exist."
        ret["result"] = False
    return ret


def enable_job(name, **kwargs):
    """
    Enable a job in the minion's schedule

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.enable_job job1
    """

    ret = {"comment": [], "result": True, "changes": {}}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = f"Job: {name} would be enabled in schedule."
    else:
        persist = kwargs.get("persist", True)

        if name in list_(show_all=True, where="opts", return_yaml=False):
            event_data = {"name": name, "func": "enable_job", "persist": persist}
        elif name in list_(show_all=True, where="pillar", return_yaml=False):
            event_data = {
                "name": name,
                "where": "pillar",
                "func": "enable_job",
                "persist": False,
            }
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](event_data, "manage_schedule")
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_enabled_job_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        # check item exists in schedule and is enabled
                        if name in schedule and schedule[name]["enabled"]:
                            ret["result"] = True
                            ret["comment"] = f"Enabled Job {name} in schedule."
                            ret["changes"][name] = "enabled"
                        else:
                            ret["result"] = False
                            ret["comment"] = f"Failed to enable job {name} in schedule."
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule enable job failed."
    return ret


def disable_job(name, **kwargs):
    """
    Disable a job in the minion's schedule

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.disable_job job1
    """

    ret = {"comment": [], "result": True, "changes": {}}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be disabled in schedule."
    else:
        persist = kwargs.get("persist", True)

        if name in list_(show_all=True, where="opts", return_yaml=False):
            event_data = {"name": name, "func": "disable_job", "persist": persist}
        elif name in list_(show_all=True, where="pillar"):
            event_data = {
                "name": name,
                "where": "pillar",
                "func": "disable_job",
                "persist": False,
            }
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](event_data, "manage_schedule")
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_disabled_job_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        # check item exists in schedule and is enabled
                        if name in schedule and not schedule[name]["enabled"]:
                            ret["result"] = True
                            ret["comment"] = f"Disabled Job {name} in schedule."
                            ret["changes"][name] = "disabled"
                        else:
                            ret["result"] = False
                            ret["comment"] = (
                                f"Failed to disable job {name} in schedule."
                            )
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule enable job failed."
    return ret


def save(**kwargs):
    """
    Save all scheduled jobs on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.save
    """

    ret = {"comment": [], "result": True}

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = "Schedule would be saved."
    else:
        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](
                    {"func": "save_schedule"}, "manage_schedule"
                )
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_saved",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        ret["result"] = True
                        ret["comment"] = "Schedule (non-pillar items) saved."
                    else:
                        ret["result"] = False
                        ret["comment"] = "Failed to save schedule."
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule save failed."
    return ret


def enable(**kwargs):
    """
    Enable all scheduled jobs on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.enable
    """

    ret = {"comment": [], "changes": {}, "result": True}

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = "Schedule would be enabled."
    else:
        persist = kwargs.get("persist", True)

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](
                    {"func": "enable", "persist": persist}, "manage_schedule"
                )
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_enabled_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        if "enabled" in schedule and schedule["enabled"]:
                            ret["result"] = True
                            ret["comment"] = "Enabled schedule on minion."
                            ret["changes"]["schedule"] = "enabled"
                        else:
                            ret["result"] = False
                            ret["comment"] = "Failed to enable schedule on minion."
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule enable job failed."
    return ret


def disable(**kwargs):
    """
    Disable all scheduled jobs on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.disable
    """

    ret = {"comment": [], "changes": {}, "result": True}

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = "Schedule would be disabled."
    else:
        persist = kwargs.get("persist", True)

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](
                    {"func": "disable", "persist": persist}, "manage_schedule"
                )
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_disabled_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        if "enabled" in schedule and not schedule["enabled"]:
                            ret["result"] = True
                            ret["comment"] = "Disabled schedule on minion."
                            ret["changes"]["schedule"] = "disabled"
                        else:
                            ret["result"] = False
                            ret["comment"] = "Failed to disable schedule on minion."
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule disable job failed."
    return ret


def reload_():
    """
    Reload saved scheduled jobs on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.reload
    """

    ret = {"comment": [], "result": True}

    # If there a schedule defined in pillar, refresh it.
    if "schedule" in __pillar__:
        out = __salt__["event.fire"]({}, "pillar_refresh")
        if out:
            ret["comment"].append("Reloaded schedule from pillar on minion.")
        else:
            ret["comment"].append("Failed to reload schedule from pillar on minion.")
            ret["result"] = False

    # move this file into an configurable opt
    sfn = "{}/{}/schedule.conf".format(
        __opts__["config_dir"], os.path.dirname(__opts__["default_include"])
    )
    if os.path.isfile(sfn):
        with salt.utils.files.fopen(sfn, "rb") as fp_:
            try:
                schedule = salt.utils.yaml.safe_load(fp_)
            except salt.utils.yaml.YAMLError as exc:
                ret["comment"].append(f"Unable to read existing schedule file: {exc}")

        if schedule:
            if "schedule" in schedule and schedule["schedule"]:
                out = __salt__["event.fire"](
                    {"func": "reload", "schedule": schedule}, "manage_schedule"
                )
                if out:
                    ret["comment"].append(
                        "Reloaded schedule on minion from schedule.conf."
                    )
                else:
                    ret["comment"].append(
                        "Failed to reload schedule on minion from schedule.conf."
                    )
                    ret["result"] = False
            else:
                ret["comment"].append(
                    "Failed to reload schedule on minion.  Saved file is empty or"
                    " invalid."
                )
                ret["result"] = False
        else:
            ret["comment"].append(
                "Failed to reload schedule on minion.  Saved file is empty or invalid."
            )
            ret["result"] = False
    return ret


def move(name, target, **kwargs):
    """
    Move scheduled job to another minion or minions.

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.move jobname target
    """

    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be moved from schedule."
    else:
        opts_schedule = list_(show_all=True, where="opts", return_yaml=False)
        pillar_schedule = list_(show_all=True, where="pillar", return_yaml=False)

        if name in opts_schedule:
            schedule_data = opts_schedule[name]
            where = None
        elif name in pillar_schedule:
            schedule_data = pillar_schedule[name]
            where = "pillar"
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        schedule_opts = []
        for key, value in schedule_data.items():
            temp = f"{key}={value}"
            schedule_opts.append(temp)
        response = __salt__["publish.publish"](target, "schedule.add", schedule_opts)

        # Get errors and list of affeced minions
        errors = []
        minions = []
        for minion in response:
            minions.append(minion)
            if not response[minion]:
                errors.append(minion)

        # parse response
        if not response:
            ret["comment"] = "no servers answered the published schedule.add command"
            return ret
        elif errors:
            ret["comment"] = "the following minions return False"
            ret["minions"] = errors
            return ret
        else:
            delete(name, where=where)
            ret["result"] = True
            ret["comment"] = f"Moved Job {name} from schedule."
            ret["minions"] = minions
            return ret
    return ret


def copy(name, target, **kwargs):
    """
    Copy scheduled job to another minion or minions.

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.copy jobname target
    """

    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    if "test" in kwargs and kwargs["test"]:
        ret["comment"] = f"Job: {name} would be copied from schedule."
    else:
        opts_schedule = list_(show_all=True, where="opts", return_yaml=False)
        pillar_schedule = list_(show_all=True, where="pillar", return_yaml=False)

        if name in opts_schedule:
            schedule_data = opts_schedule[name]
        elif name in pillar_schedule:
            schedule_data = pillar_schedule[name]
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        schedule_opts = []
        for key, value in schedule_data.items():
            temp = f"{key}={value}"
            schedule_opts.append(temp)
        response = __salt__["publish.publish"](target, "schedule.add", schedule_opts)

        # Get errors and list of affeced minions
        errors = []
        minions = []
        for minion in response:
            minions.append(minion)
            if not response[minion]:
                errors.append(minion)

        # parse response
        if not response:
            ret["comment"] = "no servers answered the published schedule.add command"
            return ret
        elif errors:
            ret["comment"] = "the following minions return False"
            ret["minions"] = errors
            return ret
        else:
            ret["result"] = True
            ret["comment"] = f"Copied Job {name} from schedule to minion(s)."
            ret["minions"] = minions
            return ret
    return ret


def postpone_job(name, current_time, new_time, **kwargs):
    """
    Postpone a job in the minion's schedule

    Current time and new time should be in date string format,
    default value is %Y-%m-%dT%H:%M:%S.

    .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.postpone_job job current_time new_time

        salt '*' schedule.postpone_job job current_time new_time time_fmt='%Y-%m-%dT%H:%M:%S'
    """

    time_fmt = kwargs.get("time_fmt") or "%Y-%m-%dT%H:%M:%S"
    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False
        return ret

    if not current_time:
        ret["comment"] = "Job current time is required."
        ret["result"] = False
        return ret
    else:
        try:
            # Validate date string
            datetime.datetime.strptime(current_time, time_fmt)
        except (TypeError, ValueError):
            log.error("Date string could not be parsed: %s, %s", new_time, time_fmt)

            ret["comment"] = "Date string could not be parsed."
            ret["result"] = False
            return ret

    if not new_time:
        ret["comment"] = "Job new_time is required."
        ret["result"] = False
        return ret
    else:
        try:
            # Validate date string
            datetime.datetime.strptime(new_time, time_fmt)
        except (TypeError, ValueError):
            log.error("Date string could not be parsed: %s, %s", new_time, time_fmt)

            ret["comment"] = "Date string could not be parsed."
            ret["result"] = False
            return ret

    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = f"Job: {name} would be postponed in schedule."
    else:

        if name in list_(show_all=True, where="opts", return_yaml=False):
            event_data = {
                "name": name,
                "time": current_time,
                "new_time": new_time,
                "time_fmt": time_fmt,
                "func": "postpone_job",
            }
        elif name in list_(show_all=True, where="pillar", return_yaml=False):
            event_data = {
                "name": name,
                "time": current_time,
                "new_time": new_time,
                "time_fmt": time_fmt,
                "where": "pillar",
                "func": "postpone_job",
            }
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](event_data, "manage_schedule")
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_postpone_job_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        # check item exists in schedule and is enabled
                        if name in schedule and schedule[name]["enabled"]:
                            ret["result"] = True
                            ret["comment"] = "Postponed Job {} in schedule.".format(
                                name
                            )
                        else:
                            ret["result"] = False
                            ret["comment"] = (
                                f"Failed to postpone job {name} in schedule."
                            )
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule postpone job failed."
    return ret


def skip_job(name, current_time, **kwargs):
    """
    Skip a job in the minion's schedule at specified time.

    Time to skip should be specified as date string format,
    default value is %Y-%m-%dT%H:%M:%S.

    .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.skip_job job time
    """
    time_fmt = kwargs.get("time_fmt") or "%Y-%m-%dT%H:%M:%S"

    ret = {"comment": [], "result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    if not current_time:
        ret["comment"] = "Job time is required."
        ret["result"] = False
    else:
        # Validate date string
        try:
            datetime.datetime.strptime(current_time, time_fmt)
        except (TypeError, ValueError):
            log.error("Date string could not be parsed: %s, %s", current_time, time_fmt)

            ret["comment"] = "Date string could not be parsed."
            ret["result"] = False
            return ret

    if "test" in __opts__ and __opts__["test"]:
        ret["comment"] = f"Job: {name} would be skipped in schedule."
    else:

        if name in list_(show_all=True, where="opts", return_yaml=False):
            event_data = {
                "name": name,
                "time": current_time,
                "time_fmt": time_fmt,
                "func": "skip_job",
            }
        elif name in list_(show_all=True, where="pillar", return_yaml=False):
            event_data = {
                "name": name,
                "time": current_time,
                "time_fmt": time_fmt,
                "where": "pillar",
                "func": "skip_job",
            }
        else:
            ret["comment"] = f"Job {name} does not exist."
            ret["result"] = False
            return ret

        try:
            with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
                res = __salt__["event.fire"](event_data, "manage_schedule")
                if res:
                    event_ret = event_bus.get_event(
                        tag="/salt/minion/minion_schedule_skip_job_complete",
                        wait=30,
                    )
                    if event_ret and event_ret["complete"]:
                        schedule = event_ret["schedule"]
                        # check item exists in schedule and is enabled
                        if name in schedule and schedule[name]["enabled"]:
                            ret["result"] = True
                            ret["comment"] = "Added Skip Job {} in schedule.".format(
                                name
                            )
                        else:
                            ret["result"] = False
                            ret["comment"] = f"Failed to skip job {name} in schedule."
                        return ret
        except KeyError:
            # Effectively a no-op, since we can't really return without an event system
            ret["comment"] = "Event module not available. Schedule skip job failed."
    return ret


def show_next_fire_time(name, **kwargs):
    """
    Show the next fire time for scheduled job

    .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.show_next_fire_time job_name

    """

    ret = {"result": True}

    if not name:
        ret["comment"] = "Job name is required."
        ret["result"] = False

    try:
        event_data = {"name": name, "func": "get_next_fire_time"}
        with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
            res = __salt__["event.fire"](event_data, "manage_schedule")
            if res:
                event_ret = event_bus.get_event(
                    tag="/salt/minion/minion_schedule_next_fire_time_complete",
                    wait=30,
                )
    except KeyError:
        # Effectively a no-op, since we can't really return without an event system
        ret = {}
        ret["comment"] = (
            "Event module not available. Schedule show next fire time failed."
        )
        ret["result"] = True
        return ret

    if "next_fire_time" in event_ret:
        ret["next_fire_time"] = event_ret["next_fire_time"]
    else:
        ret["comment"] = "next fire time not available."
    return ret


def job_status(name, time_fmt="%Y-%m-%dT%H:%M:%S"):
    """
    Show the information for a particular job.

    CLI Example:

    .. code-block:: bash

        salt '*' schedule.job_status job_name

    """

    def convert_datetime_objects_in_dict_to_string(data_dict, time_fmt):
        return {
            key: (
                value.strftime(time_fmt)
                if isinstance(value, datetime.datetime)
                else value
            )
            for key, value in data_dict.items()
        }

    schedule = {}
    try:
        with salt.utils.event.get_event("minion", opts=__opts__) as event_bus:
            res = __salt__["event.fire"](
                {"func": "job_status", "name": name, "fire_event": True},
                "manage_schedule",
            )
            if res:
                event_ret = event_bus.get_event(
                    tag="/salt/minion/minion_schedule_job_status_complete", wait=30
                )
                data = event_ret.get("data", {})
                return convert_datetime_objects_in_dict_to_string(data, time_fmt)
    except KeyError:
        # Effectively a no-op, since we can't really return without an event system
        ret = {}
        ret["comment"] = "Event module not available. Schedule list failed."
        ret["result"] = True
        log.debug("Event module not available. Schedule list failed.")
        return ret



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging

import pytest  # pylint: disable=unused-import

from salt.utils.decorators import state as statedecorators

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.core_test,
]


def test_state_output_check_changes_is_dict():
    """
    Test that changes key contains a dictionary.
    :return:
    """
    data = {"changes": []}
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert "'Changes' should be a dictionary" in out["comment"]
    assert not out["result"]


def test_state_output_check_return_is_dict():
    """
    Test for the entire return is a dictionary
    :return:
    """
    data = ["whatever"]
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert "Malformed state return. Data must be a dictionary type" in out["comment"]
    assert not out["result"]


def test_state_output_check_return_has_nrc():
    """
    Test for name/result/comment keys are inside the return.
    :return:
    """
    data = {"arbitrary": "data", "changes": {}}
    out = statedecorators.OutputUnifier("content_check")(lambda: data)()
    assert (
        " The following keys were not present in the state return: name, result, comment"
        in out["comment"]
    )
    assert not out["result"]


def test_state_output_unifier_comment_is_not_list():
    """
    Test for output is unified so the comment is converted to a multi-line string
    :return:
    """
    data = {
        "comment": ["data", "in", "the", "list"],
        "changes": {},
        "name": None,
        "result": "fantastic!",
    }
    expected = {
        "comment": "data\nin\nthe\nlist",
        "changes": {},
        "name": None,
        "result": True,
    }
    assert statedecorators.OutputUnifier("unify")(lambda: data)() == expected

    data = {
        "comment": ["data", "in", "the", "list"],
        "changes": {},
        "name": None,
        "result": None,
    }
    expected = "data\nin\nthe\nlist"
    assert statedecorators.OutputUnifier("unify")(lambda: data)()["comment"] == expected


def test_state_output_unifier_result_converted_to_true():
    """
    Test for output is unified so the result is converted to True
    :return:
    """
    data = {
        "comment": ["data", "in", "the", "list"],
        "changes": {},
        "name": None,
        "result": "Fantastic",
    }
    assert statedecorators.OutputUnifier("unify")(lambda: data)()["result"] is True


def test_state_output_unifier_result_converted_to_false():
    """
    Test for output is unified so the result is converted to False
    :return:
    """
    data = {
        "comment": ["data", "in", "the", "list"],
        "changes": {},
        "name": None,
        "result": "",
    }
    assert statedecorators.OutputUnifier("unify")(lambda: data)()["result"] is False



import os
import pathlib
import sys

_repo_root = pathlib.Path(__file__).resolve().parent.parent

if os.environ.get("ONEDIR_TESTRUN", "0") == "1":
    # In this particular case, we want to make sure that the repo root
    # is not part if sys.path so that when we import salt, we import salt from
    # the onedir and not the code checkout
    for path in list(sys.path):
        if path == "":
            sys.path.remove(path)
        elif pathlib.Path(path).resolve() == _repo_root:
            sys.path.remove(path)



"""
Module for configuring DNS Client on Windows systems
"""


def __virtual__():
    """
    Load if the module win_dns_client is loaded
    """
    if "win_dns_client.add_dns" in __salt__:
        return "win_dns_client"
    return (False, "win_dns_client module could not be loaded")


def dns_exists(name, servers=None, interface="Local Area Connection", replace=False):
    """
    Configure the DNS server list in the specified interface

    Example:

    .. code-block:: yaml

        config_dns_servers:
          win_dns_client.dns_exists:
            - replace: True #remove any servers not in the "servers" list, default is False
            - servers:
              - 8.8.8.8
              - 8.8.8.9
    """
    ret = {
        "name": name,
        "result": True,
        "changes": {
            "Servers Reordered": [],
            "Servers Added": [],
            "Servers Removed": [],
        },
        "comment": "",
    }

    if __opts__["test"]:
        ret["comment"] = "DNS Servers are set to be updated"
        ret["result"] = None
    else:
        ret["comment"] = "DNS Servers have been updated"

    # Validate syntax
    if not isinstance(servers, list):
        ret["result"] = False
        ret["comment"] = "servers entry is not a list !"
        return ret

    # Do nothing is already configured
    configured_list = __salt__["win_dns_client.get_dns_servers"](interface)
    if configured_list == servers:
        ret["comment"] = f"{servers} are already configured"
        ret["changes"] = {}
        ret["result"] = True
        return ret

    # add the DNS servers
    for i, server in enumerate(servers):
        if __opts__["test"]:
            if server in configured_list:
                if configured_list.index(server) != i:
                    ret["changes"]["Servers Reordered"].append(server)
            else:
                ret["changes"]["Servers Added"].append(server)
        else:
            if not __salt__["win_dns_client.add_dns"](server, interface, i + 1):
                ret["comment"] = "Failed to add {} as DNS server number {}".format(
                    server, i + 1
                )
                ret["result"] = False
                ret["changes"] = {}
                return ret
            else:
                if server in configured_list:
                    if configured_list.index(server) != i:
                        ret["changes"]["Servers Reordered"].append(server)
                else:
                    ret["changes"]["Servers Added"].append(server)

    # remove dns servers that weren't in our list
    if replace:
        for i, server in enumerate(configured_list):
            if server not in servers:
                if __opts__["test"]:
                    ret["changes"]["Servers Removed"].append(server)
                else:
                    if not __salt__["win_dns_client.rm_dns"](server, interface):
                        ret["comment"] = (
                            f"Failed to remove {server} from DNS server list"
                        )
                        ret["result"] = False
                        return ret
                    else:
                        ret["changes"]["Servers Removed"].append(server)

    return ret


def dns_dhcp(name, interface="Local Area Connection"):
    """
    Configure the DNS server list from DHCP Server
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    # Check the config
    config = __salt__["win_dns_client.get_dns_config"](interface)
    if config == "dhcp":
        ret["comment"] = f"{interface} already configured with DNS from DHCP"
        return ret
    else:
        ret["changes"] = {"dns": "configured from DHCP"}

    if __opts__["test"]:
        ret["result"] = None
        return ret

    # change the configuration
    ret["result"] = __salt__["win_dns_client.dns_dhcp"](interface)
    if not ret["result"]:
        ret["changes"] = {}
        ret["comment"] = 'Could not configure "{}" DNS servers from DHCP'.format(
            interface
        )

    return ret


def primary_suffix(name, suffix=None, updates=False):
    """
    .. versionadded:: 2014.7.0

    Configure the global primary DNS suffix of a DHCP client.

    suffix : None
        The suffix which is advertised for this client when acquiring a DHCP lease
        When none is set, the explicitly configured DNS suffix will be removed.

    updates : False
        Allow syncing the DNS suffix with the AD domain when the client's AD domain membership changes

    .. code-block:: yaml

        primary_dns_suffix:
            win_dns_client.primary_suffix:
                - suffix: sub.domain.tld
                - updates: True
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": "No changes needed"}

    suffix = str(suffix)

    if not isinstance(updates, bool):
        ret["result"] = False
        ret["comment"] = "'updates' must be a boolean value"
        return ret

    # TODO: waiting for an implementation of
    # https://github.com/saltstack/salt/issues/6792 to be able to handle the
    # requirement for a reboot to actually apply this state.
    # Until then, this method will only be able to verify that the required
    # value has been written to the registry and rebooting needs to be handled
    # manually

    reg_data = {
        "suffix": {
            "hive": "HKEY_LOCAL_MACHINE",
            "key": r"SYSTEM\CurrentControlSet\services\Tcpip\Parameters",
            "vname": "NV Domain",
            "vtype": "REG_SZ",
            "old": None,
            "new": suffix,
        },
        "updates": {
            "hive": "HKEY_LOCAL_MACHINE",
            "key": r"SYSTEM\CurrentControlSet\services\Tcpip\Parameters",
            "vname": "SyncDomainWithMembership",
            "vtype": "REG_DWORD",
            "old": None,
            "new": updates,
        },
    }

    reg_data["suffix"]["old"] = __utils__["reg.read_value"](
        reg_data["suffix"]["hive"],
        reg_data["suffix"]["key"],
        reg_data["suffix"]["vname"],
    )["vdata"]

    reg_data["updates"]["old"] = bool(
        __utils__["reg.read_value"](
            reg_data["updates"]["hive"],
            reg_data["updates"]["key"],
            reg_data["updates"]["vname"],
        )["vdata"]
    )

    updates_operation = "enabled" if reg_data["updates"]["new"] else "disabled"

    # No changes to suffix needed
    if reg_data["suffix"]["new"] == reg_data["suffix"]["old"]:
        # No changes to updates policy needed
        if reg_data["updates"]["new"] == reg_data["updates"]["old"]:
            return ret
        # Changes to update policy needed
        else:
            ret["comment"] = f"{updates_operation} suffix updates"
            ret["changes"] = {
                "old": {"updates": reg_data["updates"]["old"]},
                "new": {"updates": reg_data["updates"]["new"]},
            }
    # Changes to suffix needed
    else:
        # Changes to updates policy needed
        if reg_data["updates"]["new"] != reg_data["updates"]["old"]:
            ret["comment"] = (
                "Updated primary DNS suffix ({}) and {} suffix updates".format(
                    suffix, updates_operation
                )
            )
            ret["changes"] = {
                "old": {
                    "suffix": reg_data["suffix"]["old"],
                    "updates": reg_data["updates"]["old"],
                },
                "new": {
                    "suffix": reg_data["suffix"]["new"],
                    "updates": reg_data["updates"]["new"],
                },
            }
        # No changes to updates policy needed
        else:
            ret["comment"] = f"Updated primary DNS suffix ({suffix})"
            ret["changes"] = {
                "old": {"suffix": reg_data["suffix"]["old"]},
                "new": {"suffix": reg_data["suffix"]["new"]},
            }

    suffix_result = __utils__["reg.set_value"](
        reg_data["suffix"]["hive"],
        reg_data["suffix"]["key"],
        reg_data["suffix"]["vname"],
        reg_data["suffix"]["new"],
        reg_data["suffix"]["vtype"],
    )

    updates_result = __utils__["reg.set_value"](
        reg_data["updates"]["hive"],
        reg_data["updates"]["key"],
        reg_data["updates"]["vname"],
        reg_data["updates"]["new"],
        reg_data["updates"]["vtype"],
    )

    ret["result"] = suffix_result & updates_result

    return ret



"""
Tests for salt.modules.zfs

:codeauthor:    Nitin Madhok <nmadhok@g.clemson.edu>, Jorge Schrauwen <sjorge@blackdot.be>
:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs
:platform:      illumos,freebsd,linux
"""

import pytest

import salt.loader
import salt.modules.zfs as zfs
import salt.utils.zfs
from salt.utils.dateutils import strftime
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, patch
from tests.support.zfs import ZFSMockData

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture
def utils_patch():
    return ZFSMockData().get_patched_utils()


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(
        minion_opts, whitelist=["zfs", "args", "systemd", "path", "platform"]
    )
    zfs_obj = {zfs: {"__opts__": minion_opts, "__utils__": utils}}
    return zfs_obj


def test_exists_success(utils_patch):
    """
    Tests successful return of exists function
    """
    ret = {}
    ret["stdout"] = (
        "NAME        USED  AVAIL  REFER  MOUNTPOINT\nmyzpool/mydataset    30K  "
        " 157G    30K  /myzpool/mydataset"
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert zfs.exists("myzpool/mydataset")


def test_exists_failure_not_exists(utils_patch):
    """
    Tests unsuccessful return of exists function if dataset does not exist
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'myzpool/mydataset': dataset does not exist"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert not zfs.exists("myzpool/mydataset")


def test_exists_failure_invalid_name(utils_patch):
    """
    Tests unsuccessful return of exists function if dataset name is invalid
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'myzpool/': invalid dataset name"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert not zfs.exists("myzpool/")


def test_create_success(utils_patch):
    """
    Tests successful return of create function on ZFS file system creation
    """
    res = OrderedDict([("created", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool/mydataset")


def test_create_success_with_create_parent(utils_patch):
    """
    Tests successful return of create function when ``create_parent=True``
    """
    res = OrderedDict([("created", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool/mydataset/mysubdataset", create_parent=True)


def test_create_success_with_properties(utils_patch):
    """
    Tests successful return of create function on ZFS file system creation (with properties)
    """
    res = OrderedDict([("created", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create(
            "myzpool/mydataset",
            properties={"mountpoint": "/export/zfs", "sharenfs": "on"},
        )


def test_create_error_missing_dataset(utils_patch):
    """
    Tests unsuccessful return of create function if dataset name is missing
    """
    res = OrderedDict(
        [
            ("created", False),
            ("error", "cannot create 'myzpool': missing dataset name"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot create 'myzpool': missing dataset name"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool")


def test_create_error_trailing_slash(utils_patch):
    """
    Tests unsuccessful return of create function if trailing slash in name is present
    """
    res = OrderedDict(
        [
            ("created", False),
            ("error", "cannot create 'myzpool/': trailing slash in name"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot create 'myzpool/': trailing slash in name"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool/")


def test_create_error_no_such_pool(utils_patch):
    """
    Tests unsuccessful return of create function if the pool is not present
    """
    res = OrderedDict(
        [
            ("created", False),
            ("error", "cannot create 'myzpool/mydataset': no such pool 'myzpool'"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot create 'myzpool/mydataset': no such pool 'myzpool'"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool/mydataset")


def test_create_error_missing_parent(utils_patch):
    """
    Tests unsuccessful return of create function if the parent datasets do not exist
    """
    res = OrderedDict(
        [
            ("created", False),
            (
                "error",
                "cannot create 'myzpool/mydataset/mysubdataset': parent does not"
                " exist",
            ),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = (
        "cannot create 'myzpool/mydataset/mysubdataset': parent does not exist"
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.create("myzpool/mydataset/mysubdataset")


def test_destroy_success(utils_patch):
    """
    Tests successful return of destroy function on ZFS file system destruction
    """
    res = OrderedDict([("destroyed", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.destroy("myzpool/mydataset")


def test_destroy_error_not_exists(utils_patch):
    """
    Tests failure return of destroy function on ZFS file system destruction
    """
    res = OrderedDict(
        [
            ("destroyed", False),
            ("error", "cannot open 'myzpool/mydataset': dataset does not exist"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'myzpool/mydataset': dataset does not exist"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.destroy("myzpool/mydataset")


def test_destroy_error_has_children(utils_patch):
    """
    Tests failure return of destroy function on ZFS file system destruction
    """
    res = OrderedDict(
        [
            ("destroyed", False),
            (
                "error",
                "\n".join(
                    [
                        "cannot destroy 'myzpool/mydataset': filesystem has"
                        " children",
                        "use 'recursive=True' to destroy the following datasets:",
                        "myzpool/mydataset@snapshot",
                    ]
                ),
            ),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "\n".join(
        [
            "cannot destroy 'myzpool/mydataset': filesystem has children",
            "use '-r' to destroy the following datasets:",
            "myzpool/mydataset@snapshot",
        ]
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.destroy("myzpool/mydataset")


@pytest.mark.slow_test
def test_rename_success(utils_patch):
    """
    Tests successful return of rename function
    """
    res = OrderedDict([("renamed", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.rename("myzpool/mydataset", "myzpool/newdataset")


def test_rename_error_not_exists(utils_patch):
    """
    Tests failure return of rename function
    """
    res = OrderedDict(
        [
            ("renamed", False),
            ("error", "cannot open 'myzpool/mydataset': dataset does not exist"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'myzpool/mydataset': dataset does not exist"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.rename("myzpool/mydataset", "myzpool/newdataset")


def test_list_success(utils_patch):
    """
    Tests zfs list
    """
    res = OrderedDict(
        [
            (
                "myzpool",
                OrderedDict(
                    [
                        ("used", 849329782784),
                        ("avail", 1081258016768),
                        ("refer", 98304),
                        ("mountpoint", "/myzpool"),
                    ]
                ),
            ),
        ]
    )
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "myzpool\t791G\t1007G\t96K\t/myzpool"
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_("myzpool")


@pytest.mark.slow_test
def test_list_parsable_success(utils_patch):
    """
    Tests zfs list with parsable set to False
    """
    res = OrderedDict(
        [
            (
                "myzpool",
                OrderedDict(
                    [
                        ("used", "791G"),
                        ("avail", "1007G"),
                        ("refer", "96K"),
                        ("mountpoint", "/myzpool"),
                    ]
                ),
            ),
        ]
    )
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "myzpool\t791G\t1007G\t96K\t/myzpool"
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_("myzpool", parsable=False)


def test_list_custom_success(utils_patch):
    """
    Tests zfs list
    """
    res = OrderedDict(
        [
            (
                "myzpool",
                OrderedDict(
                    [
                        ("canmount", True),
                        ("used", 849329782784),
                        ("avail", 1081258016768),
                        ("compression", False),
                    ]
                ),
            ),
        ]
    )
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "myzpool\ton\t791G\t1007G\toff"
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_("myzpool", properties="canmount,used,avail,compression")


def test_list_custom_parsable_success(utils_patch):
    """
    Tests zfs list
    """
    res = OrderedDict(
        [
            (
                "myzpool",
                OrderedDict(
                    [
                        ("canmount", "on"),
                        ("used", "791G"),
                        ("avail", "1007G"),
                        ("compression", "off"),
                    ]
                ),
            ),
        ]
    )
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "myzpool\ton\t791G\t1007G\toff"
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_(
            "myzpool",
            properties="canmount,used,avail,compression",
            parsable=False,
        )


def test_list_error_no_dataset(utils_patch):
    """
    Tests zfs list
    """
    res = OrderedDict()
    ret = {}
    ret["retcode"] = 1
    ret["stdout"] = "cannot open 'myzpool': dataset does not exist"
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_("myzpool")


@pytest.mark.slow_test
def test_list_mount_success(utils_patch):
    """
    Tests zfs list_mount
    """
    res = OrderedDict([("myzpool/data", "/data"), ("myzpool/data/ares", "/data/ares")])
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "\n".join(
        ["myzpool/data\t\t\t\t/data", "myzpool/data/ares\t\t\t/data/ares"]
    )
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.list_mount()


def test_mount_success(utils_patch):
    """
    Tests zfs mount of filesystem
    """
    res = OrderedDict([("mounted", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.mount("myzpool/mydataset")


def test_mount_failure(utils_patch):
    """
    Tests zfs mount of already mounted filesystem
    """
    res = OrderedDict(
        [
            ("mounted", False),
            (
                "error",
                "cannot mount 'myzpool/mydataset': filesystem already mounted",
            ),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot mount 'myzpool/mydataset': filesystem already mounted"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.mount("myzpool/mydataset")


def test_unmount_success(utils_patch):
    """
    Tests zfs unmount of filesystem
    """
    res = OrderedDict([("unmounted", True)])
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.unmount("myzpool/mydataset")


def test_unmount_failure(utils_patch):
    """
    Tests zfs unmount of already mounted filesystem
    """
    res = OrderedDict(
        [
            ("unmounted", False),
            ("error", "cannot mount 'myzpool/mydataset': not currently mounted"),
        ]
    )
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot mount 'myzpool/mydataset': not currently mounted"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.unmount("myzpool/mydataset")


def test_inherit_success(utils_patch):
    """
    Tests zfs inherit of compression property
    """
    res = OrderedDict([("inherited", True)])
    ret = {"pid": 45193, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.inherit("compression", "myzpool/mydataset")


def test_inherit_failure(utils_patch):
    """
    Tests zfs inherit of canmount
    """
    res = OrderedDict(
        [
            ("inherited", False),
            ("error", "'canmount' property cannot be inherited"),
        ]
    )
    ret = {
        "pid": 43898,
        "retcode": 1,
        "stderr": "'canmount' property cannot be inherited",
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.inherit("canmount", "myzpool/mydataset")


@pytest.mark.slow_test
def test_diff(utils_patch):
    """
    Tests zfs diff
    """
    res = [
        "1517063879.144517494\tM\t\t/data/test/",
        "1517063875.296592355\t+\t\t/data/test/world",
        "1517063879.274438467\t+\t\t/data/test/hello",
    ]
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "\n".join(
        [
            "1517063879.144517494\tM\t\t/data/test/",
            "1517063875.296592355\t+\t\t/data/test/world",
            "1517063879.274438467\t+\t\t/data/test/hello",
        ]
    )
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.diff("myzpool/mydataset@yesterday", "myzpool/mydataset")


def test_diff_parsed_time(utils_patch):
    """
    Tests zfs diff
    """
    ## NOTE: do not hardcode parsed timestamps, timezone play a role here.
    ##       zfs diff output seems to be timezone aware
    res = OrderedDict(
        [
            (
                strftime(1517063879.144517494, "%Y-%m-%d.%H:%M:%S.%f"),
                "M\t\t/data/test/",
            ),
            (
                strftime(1517063875.296592355, "%Y-%m-%d.%H:%M:%S.%f"),
                "+\t\t/data/test/world",
            ),
            (
                strftime(1517063879.274438467, "%Y-%m-%d.%H:%M:%S.%f"),
                "+\t\t/data/test/hello",
            ),
        ]
    )
    ret = {}
    ret["retcode"] = 0
    ret["stdout"] = "\n".join(
        [
            "1517063879.144517494\tM\t\t/data/test/",
            "1517063875.296592355\t+\t\t/data/test/world",
            "1517063879.274438467\t+\t\t/data/test/hello",
        ]
    )
    ret["stderr"] = ""
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.diff("myzpool/data@yesterday", "myzpool/data", parsable=False)


@pytest.mark.slow_test
def test_rollback_success(utils_patch):
    """
    Tests zfs rollback success
    """
    res = OrderedDict([("rolledback", True)])
    ret = {"pid": 56502, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.rollback("myzpool/mydataset@yesterday")


@pytest.mark.slow_test
def test_rollback_failure(utils_patch):
    """
    Tests zfs rollback failure
    """
    res = OrderedDict(
        [
            ("rolledback", False),
            (
                "error",
                "\n".join(
                    [
                        "cannot rollback to 'myzpool/mydataset@yesterday': more"
                        " recent snapshots or bookmarks exist",
                        "use 'recursive=True' to force deletion of the following"
                        " snapshots and bookmarks:",
                        "myzpool/mydataset@today",
                    ]
                ),
            ),
        ]
    )
    ret = {
        "pid": 57471,
        "retcode": 1,
        "stderr": (
            "cannot rollback to 'myzpool/mydataset@yesterday': more recent"
            " snapshots or bookmarks exist\nuse '-r' to force deletion of the"
            " following snapshots and bookmarks:\nmyzpool/mydataset@today"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.rollback("myzpool/mydataset@yesterday")


def test_clone_success(utils_patch):
    """
    Tests zfs clone success
    """
    res = OrderedDict([("cloned", True)])
    ret = {"pid": 64532, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.clone("myzpool/mydataset@yesterday", "myzpool/yesterday")


def test_clone_failure(utils_patch):
    """
    Tests zfs clone failure
    """
    res = OrderedDict(
        [
            ("cloned", False),
            (
                "error",
                "cannot create 'myzpool/archive/yesterday': parent does not exist",
            ),
        ]
    )
    ret = {
        "pid": 64864,
        "retcode": 1,
        "stderr": ("cannot create 'myzpool/archive/yesterday': parent does not exist"),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.clone(
            "myzpool/mydataset@yesterday", "myzpool/archive/yesterday"
        )


@pytest.mark.slow_test
def test_promote_success(utils_patch):
    """
    Tests zfs promote success
    """
    res = OrderedDict([("promoted", True)])
    ret = {"pid": 69075, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.promote("myzpool/yesterday")


def test_promote_failure(utils_patch):
    """
    Tests zfs promote failure
    """
    res = OrderedDict(
        [
            ("promoted", False),
            (
                "error",
                "cannot promote 'myzpool/yesterday': not a cloned filesystem",
            ),
        ]
    )
    ret = {
        "pid": 69209,
        "retcode": 1,
        "stderr": "cannot promote 'myzpool/yesterday': not a cloned filesystem",
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.promote("myzpool/yesterday")


@pytest.mark.slow_test
def test_bookmark_success(utils_patch):
    """
    Tests zfs bookmark success
    """
    with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/man")):
        res = OrderedDict([("bookmarked", True)])
        ret = {"pid": 20990, "retcode": 0, "stderr": "", "stdout": ""}
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
            zfs.__utils__, utils_patch
        ):
            assert res == zfs.bookmark(
                "myzpool/mydataset@yesterday", "myzpool/mydataset#important"
            )


@pytest.mark.slow_test
def test_holds_success(utils_patch):
    """
    Tests zfs holds success
    """
    res = OrderedDict(
        [
            ("important", "Wed Dec 23 21:06 2015"),
            ("release-1.0", "Wed Dec 23 21:08 2015"),
        ]
    )
    ret = {
        "pid": 40216,
        "retcode": 0,
        "stderr": "",
        "stdout": (
            "myzpool/mydataset@baseline\timportant  \tWed Dec 23 21:06"
            " 2015\nmyzpool/mydataset@baseline\trelease-1.0\tWed Dec 23 21:08 2015"
        ),
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.holds("myzpool/mydataset@baseline")


def test_holds_failure(utils_patch):
    """
    Tests zfs holds failure
    """
    res = OrderedDict(
        [
            (
                "error",
                "cannot open 'myzpool/mydataset@baseline': dataset does not exist",
            ),
        ]
    )
    ret = {
        "pid": 40993,
        "retcode": 1,
        "stderr": ("cannot open 'myzpool/mydataset@baseline': dataset does not exist"),
        "stdout": "no datasets available",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.holds("myzpool/mydataset@baseline")


def test_hold_success(utils_patch):
    """
    Tests zfs hold success
    """
    res = OrderedDict([("held", True)])
    ret = {"pid": 50876, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.hold(
            "important",
            "myzpool/mydataset@baseline",
            "myzpool/mydataset@release-1.0",
        )


def test_hold_failure(utils_patch):
    """
    Tests zfs hold failure
    """
    res = OrderedDict(
        [
            ("held", False),
            (
                "error",
                "cannot hold snapshot 'myzpool/mydataset@baseline': tag already"
                " exists on this dataset",
            ),
        ]
    )
    ret = {
        "pid": 51006,
        "retcode": 1,
        "stderr": (
            "cannot hold snapshot 'myzpool/mydataset@baseline': tag already exists"
            " on this dataset"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.hold("important", "myzpool/mydataset@baseline")


def test_release_success(utils_patch):
    """
    Tests zfs release success
    """
    res = OrderedDict([("released", True)])
    ret = {"pid": 50876, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.release(
            "important",
            "myzpool/mydataset@baseline",
            "myzpool/mydataset@release-1.0",
        )


@pytest.mark.slow_test
def test_release_failure(utils_patch):
    """
    Tests zfs release failure
    """
    res = OrderedDict(
        [
            ("released", False),
            (
                "error",
                "cannot release hold from snapshot 'myzpool/mydataset@baseline': no"
                " such tag on this dataset",
            ),
        ]
    )
    ret = {
        "pid": 51006,
        "retcode": 1,
        "stderr": (
            "cannot release hold from snapshot 'myzpool/mydataset@baseline': no"
            " such tag on this dataset"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.release("important", "myzpool/mydataset@baseline")


def test_snapshot_success(utils_patch):
    """
    Tests zfs snapshot success
    """
    res = OrderedDict([("snapshotted", True)])
    ret = {"pid": 69125, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.snapshot("myzpool/mydataset@baseline")


def test_snapshot_failure(utils_patch):
    """
    Tests zfs snapshot failure
    """
    res = OrderedDict(
        [
            ("snapshotted", False),
            (
                "error",
                "cannot create snapshot 'myzpool/mydataset@baseline': dataset"
                " already exists",
            ),
        ]
    )
    ret = {
        "pid": 68526,
        "retcode": 1,
        "stderr": (
            "cannot create snapshot 'myzpool/mydataset@baseline': dataset already"
            " exists"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.snapshot("myzpool/mydataset@baseline")


def test_snapshot_failure2(utils_patch):
    """
    Tests zfs snapshot failure
    """
    res = OrderedDict(
        [
            ("snapshotted", False),
            ("error", "cannot open 'myzpool/mydataset': dataset does not exist"),
        ]
    )
    ret = {
        "pid": 69256,
        "retcode": 2,
        "stderr": (
            "cannot open 'myzpool/mydataset': dataset does not"
            " exist\nusage:\n\tsnapshot [-r] [-o property=value] ..."
            " <filesystem|volume>@<snap> ...\n\nFor the property list, run: zfs"
            " set|get\n\nFor the delegated permission list, run: zfs allow|unallow"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.snapshot("myzpool/mydataset@baseline")


def test_set_success(utils_patch):
    """
    Tests zfs set success
    """
    res = OrderedDict([("set", True)])
    ret = {"pid": 79736, "retcode": 0, "stderr": "", "stdout": ""}
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.set("myzpool/mydataset", compression="lz4")


@pytest.mark.slow_test
def test_set_failure(utils_patch):
    """
    Tests zfs set failure
    """
    res = OrderedDict(
        [
            ("set", False),
            (
                "error",
                "cannot set property for 'myzpool/mydataset': 'canmount' must be"
                " one of 'on | off | noauto'",
            ),
        ]
    )
    ret = {
        "pid": 79887,
        "retcode": 1,
        "stderr": (
            "cannot set property for 'myzpool/mydataset': 'canmount' must be one of"
            " 'on | off | noauto'"
        ),
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.set("myzpool/mydataset", canmount="lz4")


def test_get_success(utils_patch):
    """
    Tests zfs get success
    """
    res = OrderedDict(
        [
            (
                "myzpool",
                OrderedDict([("used", OrderedDict([("value", 906238099456)]))]),
            ),
        ]
    )
    ret = {
        "pid": 562,
        "retcode": 0,
        "stderr": "",
        "stdout": "myzpool\tused\t906238099456",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.get("myzpool", properties="used", fields="value")


def test_get_parsable_success(utils_patch):
    """
    Tests zfs get with parsable output
    """
    res = OrderedDict(
        [("myzpool", OrderedDict([("used", OrderedDict([("value", "844G")]))]))]
    )
    ret = {
        "pid": 562,
        "retcode": 0,
        "stderr": "",
        "stdout": "myzpool\tused\t906238099456",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zfs.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zfs.__utils__, utils_patch
    ):
        assert res == zfs.get(
            "myzpool", properties="used", fields="value", parsable=False
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import logging
import os

import pytest

import salt.modules.mysql as mysql_mod
import salt.states.mysql_query as mysql_query
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)

MySQLdb = pytest.importorskip("MySQLdb")
pymysql = pytest.importorskip("pymysql")

pymysql.install_as_MySQLdb()


class MockMySQLConnect:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def autocommit(self, *args, **kwards):
        return True

    def cursor(self, *args, **kwards):
        return MagicMock()


@pytest.fixture
def configure_loader_modules():
    return {mysql_query: {}, mysql_mod: {}}


def test_run():
    """
    Test to execute an arbitrary query on the specified database.
    """
    name = "query_id"
    database = "my_database"
    query = "SELECT * FROM table;"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_t = MagicMock(return_value=True)
    mock_f = MagicMock(return_value=False)
    mock_str = MagicMock(return_value="salt")
    mock_none = MagicMock(return_value=None)
    mock_dict = MagicMock(return_value={"salt": "SALT"})
    mock_lst = MagicMock(return_value=["grain"])
    with patch.dict(mysql_query.__salt__, {"mysql.db_exists": mock_f}):
        with patch.object(mysql_query, "_get_mysql_error", mock_str):
            ret.update({"comment": "salt", "result": False})
            assert mysql_query.run(name, database, query) == ret

        with patch.object(mysql_query, "_get_mysql_error", mock_none):
            comt = f"Database {name} is not present"
            ret.update({"comment": comt, "result": None})
            assert mysql_query.run(name, database, query) == ret

    with patch.dict(
        mysql_query.__salt__,
        {
            "mysql.db_exists": mock_t,
            "grains.ls": mock_lst,
            "grains.get": mock_dict,
            "mysql.query": mock_str,
        },
    ):
        comt = "No execution needed. Grain grain already set"
        ret.update({"comment": comt, "result": True})
        assert (
            mysql_query.run(
                name,
                database,
                query,
                output="grain",
                grain="grain",
                overwrite=False,
            )
            == ret
        )

        with patch.dict(mysql_query.__opts__, {"test": True}):
            comt = "Query would execute, storing result in grain: grain"
            ret.update({"comment": comt, "result": None})
            assert (
                mysql_query.run(name, database, query, output="grain", grain="grain")
                == ret
            )

            comt = "Query would execute, storing result in grain: grain:salt"
            ret.update({"comment": comt})
            assert (
                mysql_query.run(
                    name, database, query, output="grain", grain="grain", key="salt"
                )
                == ret
            )

            comt = "Query would execute, storing result in file: salt"
            ret.update({"comment": comt})
            assert (
                mysql_query.run(name, database, query, output="salt", grain="grain")
                == ret
            )

            comt = "Query would execute, not storing result"
            ret.update({"comment": comt})
            assert mysql_query.run(name, database, query) == ret

        comt = "No execution needed. Grain grain:salt already set"
        ret.update({"comment": comt, "result": True})
        assert (
            mysql_query.run(
                name,
                database,
                query,
                output="grain",
                grain="grain",
                key="salt",
                overwrite=False,
            )
            == ret
        )

        comt = "Error: output type 'grain' needs the grain parameter\n"
        ret.update({"comment": comt, "result": False})
        assert mysql_query.run(name, database, query, output="grain") == ret

        with patch.object(os.path, "isfile", mock_t):
            comt = "No execution needed. File salt already set"
            ret.update({"comment": comt, "result": True})
            assert (
                mysql_query.run(
                    name,
                    database,
                    query,
                    output="salt",
                    grain="grain",
                    overwrite=False,
                )
                == ret
            )

        with patch.dict(mysql_query.__opts__, {"test": False}):
            ret.update({"comment": "salt", "changes": {"query": "Executed"}})
            assert mysql_query.run(name, database, query) == ret


def test_run_multiple_statements():
    """
    Test to execute an arbitrary query on the specified database
    and ensure that the correct multi_statements flag is passed along
    to MySQLdb.connect.
    """
    name = "query_id"
    database = "my_database"
    query = "SELECT * FROM table; SELECT * from another_table;"

    mock_t = MagicMock(return_value=True)

    with patch.dict(mysql_query.__salt__, {"mysql.db_exists": mock_t}), patch.dict(
        mysql_query.__opts__, {"test": False}
    ), patch.dict(mysql_query.__salt__, {"mysql.query": mysql_mod.query}), patch.dict(
        mysql_query.__salt__, {"mysql._execute": MagicMock()}
    ), patch.dict(
        mysql_mod.__salt__, {"config.option": MagicMock()}
    ), patch(
        "MySQLdb.connect", return_value=MockMySQLConnect()
    ) as mock_connect:
        ret = mysql_query.run(name, database, query, client_flags=["multi_statements"])
        assert 1 == len(mock_connect.mock_calls)
        assert "client_flag=65536" in str(mock_connect.mock_calls[0])



"""
The networking module for NI Linux Real-Time distro

"""

import configparser
import logging
import os
import re
import time

import salt.exceptions
import salt.utils.files
import salt.utils.validate.net

try:
    import pyconnman
except ImportError:
    pyconnman = None

try:
    import dbus
except ImportError:
    dbus = None

try:
    import pyiface
    from pyiface.ifreqioctls import IFF_LOOPBACK, IFF_RUNNING
except ImportError:
    pyiface = None

try:
    from requests.structures import CaseInsensitiveDict
except ImportError:
    CaseInsensitiveDict = None

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "ip"

SERVICE_PATH = "/net/connman/service/"
INTERFACES_CONFIG = "/var/lib/connman/interfaces.config"
NIRTCFG_PATH = "/usr/local/natinst/bin/nirtcfg"
INI_FILE = "/etc/natinst/share/ni-rt.ini"
_CONFIG_TRUE = ["yes", "on", "true", "1", True]
NIRTCFG_ETHERCAT = "EtherCAT"


def _assume_condition(condition, err):
    """
    Raise an exception if the condition is false
    """
    if not condition:
        raise RuntimeError(err)


def __virtual__():
    """
    Confine this module to NI Linux Real-Time based distros
    """
    try:
        msg = "The nilrt_ip module could not be loaded: unsupported OS family"
        _assume_condition(__grains__["os_family"] == "NILinuxRT", msg)
        _assume_condition(
            CaseInsensitiveDict, "The python package request is not installed"
        )
        _assume_condition(pyiface, "The python pyiface package is not installed")
        if __grains__["lsb_distrib_id"] != "nilrt":
            _assume_condition(
                pyconnman, "The python package pyconnman is not installed"
            )
            _assume_condition(dbus, "The python DBus package is not installed")
            _assume_condition(_get_state() != "offline", "Connman is not running")
    except RuntimeError as exc:
        return False, str(exc)
    return __virtualname__


def _get_state():
    """
    Returns the state of connman
    """
    try:
        return pyconnman.ConnManager().get_property("State")
    except KeyError:
        return "offline"
    except dbus.DBusException as exc:
        raise salt.exceptions.CommandExecutionError(f"Connman daemon error: {exc}")


def _get_technologies():
    """
    Returns the technologies of connman
    """
    tech = ""
    technologies = pyconnman.ConnManager().get_technologies()
    for path, params in technologies:
        tech += (
            "{}\n\tName = {}\n\tType = {}\n\tPowered = {}\n\tConnected = {}\n".format(
                path,
                params["Name"],
                params["Type"],
                params["Powered"] == 1,
                params["Connected"] == 1,
            )
        )
    return tech


def _get_services():
    """
    Returns a list with all connman services
    """
    serv = []
    services = pyconnman.ConnManager().get_services()
    for path, _ in services:
        serv.append(str(path[len(SERVICE_PATH) :]))
    return serv


def _connected(service):
    """
    Verify if a connman service is connected
    """
    state = pyconnman.ConnService(os.path.join(SERVICE_PATH, service)).get_property(
        "State"
    )
    return state == "online" or state == "ready"


def _space_delimited_list(value):
    """
    validate that a value contains one or more space-delimited values
    """
    if isinstance(value, str):
        items = value.split(" ")
        valid = items and all(items)
    else:
        valid = hasattr(value, "__iter__") and (value != [])

    if valid:
        return True, "space-delimited string"
    return False, f"{value} is not a valid list.\n"


def _validate_ipv4(value):
    """
    validate ipv4 values
    """
    if len(value) == 3:
        if not salt.utils.validate.net.ipv4_addr(value[0].strip()):
            return False, f"Invalid ip address: {value[0]} for ipv4 option"
        if not salt.utils.validate.net.netmask(value[1].strip()):
            return False, f"Invalid netmask: {value[1]} for ipv4 option"
        if not salt.utils.validate.net.ipv4_addr(value[2].strip()):
            return False, f"Invalid gateway: {value[2]} for ipv4 option"
    else:
        return False, f"Invalid value: {value} for ipv4 option"
    return True, ""


def _interface_to_service(iface):
    """
    returns the corresponding service to given interface if exists, otherwise return None
    """
    for _service in _get_services():
        service_info = pyconnman.ConnService(os.path.join(SERVICE_PATH, _service))
        if service_info.get_property("Ethernet")["Interface"] == iface:
            return _service
    return None


def _get_service_info(service):
    """
    return details about given connman service
    """
    service_info = pyconnman.ConnService(os.path.join(SERVICE_PATH, service))
    data = {
        "label": service,
        "wireless": service_info.get_property("Type") == "wifi",
        "connectionid": str(service_info.get_property("Ethernet")["Interface"]),
        "hwaddr": str(service_info.get_property("Ethernet")["Address"]),
    }

    state = service_info.get_property("State")
    if state == "ready" or state == "online":
        data["up"] = True
        data["ipv4"] = {"gateway": "0.0.0.0"}
        ipv4 = "IPv4"
        if service_info.get_property("IPv4")["Method"] == "manual":
            ipv4 += ".Configuration"
        ipv4_info = service_info.get_property(ipv4)
        for info in ["Method", "Address", "Netmask", "Gateway"]:
            value = ipv4_info.get(info)
            if value is None:
                log.warning("Unable to get IPv4 %s for service %s\n", info, service)
                continue
            if info == "Method":
                info = "requestmode"
                if value == "dhcp":
                    value = "dhcp_linklocal"
                elif value in ("manual", "fixed"):
                    value = "static"
            data["ipv4"][info.lower()] = str(value)

        ipv6_info = service_info.get_property("IPv6")
        for info in ["Address", "Prefix", "Gateway"]:
            value = ipv6_info.get(info)
            if value is None:
                log.warning("Unable to get IPv6 %s for service %s\n", info, service)
                continue
            data["ipv6"][info.lower()] = [str(value)]

        nameservers = []
        for nameserver_prop in service_info.get_property("Nameservers"):
            nameservers.append(str(nameserver_prop))
        data["ipv4"]["dns"] = nameservers
    else:
        data["up"] = False
        data["ipv4"] = {"requestmode": "disabled"}

    data["ipv4"]["supportedrequestmodes"] = ["dhcp_linklocal", "disabled", "static"]
    return data


def _get_dns_info():
    """
    return dns list
    """
    dns_list = []
    try:
        with salt.utils.files.fopen("/etc/resolv.conf", "r+") as dns_info:
            lines = dns_info.readlines()
            for line in lines:
                if "nameserver" in line:
                    dns = line.split()[1].strip()
                    if dns not in dns_list:
                        dns_list.append(dns)
    except OSError:
        log.warning("Could not get domain\n")
    return dns_list


def _remove_quotes(value):
    """
    Remove leading and trailing double quotes if they exist.
    """
    # nirtcfg writes values with quotes
    if len(value) > 1 and value[0] == value[-1] == '"':
        value = value[1:-1]
    return value


def _load_config(section, options, default_value="", filename=INI_FILE):
    """
    Get values for some options and a given section from a config file.

    :param section: Section Name
    :param options: List of options
    :param default_value: Default value if an option doesn't have a value. Default is empty string.
    :param filename: config file. Default is INI_FILE.
    :return:
    """
    results = {}
    if not options:
        return results
    with salt.utils.files.fopen(filename, "r") as config_file:
        config_parser = configparser.RawConfigParser(
            dict_type=CaseInsensitiveDict, converters={"unquoted": _remove_quotes}
        )
        config_parser.readfp(config_file)
        for option in options:
            results[option] = config_parser.getunquoted(
                section, option, fallback=default_value
            )

    return results


def _get_request_mode_info(interface):
    """
    return requestmode for given interface
    """
    settings = _load_config(interface, ["linklocalenabled", "dhcpenabled"], -1)
    link_local_enabled = int(settings["linklocalenabled"])
    dhcp_enabled = int(settings["dhcpenabled"])

    if dhcp_enabled == 1:
        return "dhcp_linklocal" if link_local_enabled == 1 else "dhcp_only"
    else:
        if link_local_enabled == 1:
            return "linklocal_only"
        if link_local_enabled == 0:
            return "static"

    # some versions of nirtcfg don't set the dhcpenabled/linklocalenabled variables
    # when selecting "DHCP or Link Local" from MAX, so return it by default to avoid
    # having the requestmode "None" because none of the conditions above matched.
    return "dhcp_linklocal"


def _get_adapter_mode_info(interface):
    """
    return adaptermode for given interface
    """
    mode = _load_config(interface, ["mode"])["mode"].lower()
    return mode if mode in ["disabled", "ethercat"] else "tcpip"


def _get_possible_adapter_modes(interface, blacklist):
    """
    Return possible adapter modes for a given interface using a blacklist.

    :param interface: interface name
    :param blacklist: given blacklist
    :return: list of possible adapter modes
    """
    adapter_modes = []
    protocols = _load_config("lvrt", ["AdditionalNetworkProtocols"])[
        "AdditionalNetworkProtocols"
    ].lower()
    sys_interface_path = os.readlink(f"/sys/class/net/{interface}")
    with salt.utils.files.fopen(f"/sys/class/net/{interface}/uevent") as uevent_file:
        uevent_lines = uevent_file.readlines()
    uevent_devtype = ""
    for line in uevent_lines:
        if line.startswith("DEVTYPE="):
            uevent_devtype = line.split("=")[1].strip()
            break

    for adapter_mode in blacklist:
        if adapter_mode == "_":
            continue
        value = blacklist.get(adapter_mode, {})
        if value.get("additional_protocol") and adapter_mode not in protocols:
            continue

        if interface not in value["name"] and not any(
            (blacklist["_"][iface_type] == "sys" and iface_type in sys_interface_path)
            or (blacklist["_"][iface_type] == "uevent" and iface_type == uevent_devtype)
            for iface_type in value["type"]
        ):
            adapter_modes += [adapter_mode]
    return adapter_modes


def _get_static_info(interface):
    """
    Return information about an interface from config file.

    :param interface: interface label
    """
    data = {
        "connectionid": interface.name,
        "label": interface.name,
        "hwaddr": interface.hwaddr[:-1],
        "up": False,
        "ipv4": {
            "supportedrequestmodes": ["dhcp_linklocal", "disabled", "static"],
            "requestmode": "dhcp_linklocal",
        },
        "wireless": False,
    }
    hwaddr_section_number = "".join(data["hwaddr"].split(":"))
    if os.path.exists(INTERFACES_CONFIG):
        information = _load_config(
            "service_" + hwaddr_section_number,
            ["IPv4", "Nameservers", "IPv4.method"],
            filename=INTERFACES_CONFIG,
        )
        if information["IPv4.method"] == "manual" and information["IPv4"] != "":
            ipv4_information = information["IPv4"].split("/")
            data["ipv4"]["address"] = ipv4_information[0]
            data["ipv4"]["dns"] = (
                ""
                if information["Nameservers"] == "''"
                else information["Nameservers"].split(",")
            )
            data["ipv4"]["netmask"] = ipv4_information[1]
            data["ipv4"]["gateway"] = ipv4_information[2]
            data["ipv4"]["requestmode"] = "static"
        elif information["IPv4"] == "off":
            data["ipv4"]["requestmode"] = "disabled"

    return data


def _get_base_interface_info(interface):
    """
    return base details about given interface
    """
    blacklist = {
        "tcpip": {"name": [], "type": [], "additional_protocol": False},
        "disabled": {
            "name": ["eth0"],
            "type": ["gadget"],
            "additional_protocol": False,
        },
        "ethercat": {
            "name": ["eth0"],
            "type": ["gadget", "usb", "wlan"],
            "additional_protocol": True,
        },
        "_": {"usb": "sys", "gadget": "uevent", "wlan": "uevent"},
    }
    return {
        "label": interface.name,
        "connectionid": interface.name,
        "supported_adapter_modes": _get_possible_adapter_modes(
            interface.name, blacklist
        ),
        "adapter_mode": _get_adapter_mode_info(interface.name),
        "up": interface.flags & IFF_RUNNING != 0,
        "ipv4": {
            "supportedrequestmodes": [
                "dhcp_linklocal",
                "dhcp_only",
                "linklocal_only",
                "static",
            ],
            "requestmode": _get_request_mode_info(interface.name),
        },
        "hwaddr": interface.hwaddr[:-1],
    }


def _get_ethercat_interface_info(interface):
    """
    return details about given ethercat interface
    """
    base_information = _get_base_interface_info(interface)
    base_information["ethercat"] = {
        "masterid": _load_config(interface.name, ["MasterID"])["MasterID"]
    }
    return base_information


def _get_tcpip_interface_info(interface):
    """
    return details about given tcpip interface
    """
    base_information = _get_base_interface_info(interface)
    if base_information["ipv4"]["requestmode"] == "static":
        settings = _load_config(
            interface.name, ["IP_Address", "Subnet_Mask", "Gateway", "DNS_Address"]
        )
        base_information["ipv4"]["address"] = settings["IP_Address"]
        base_information["ipv4"]["netmask"] = settings["Subnet_Mask"]
        base_information["ipv4"]["gateway"] = settings["Gateway"]
        base_information["ipv4"]["dns"] = [settings["DNS_Address"]]
    elif base_information["up"]:
        base_information["ipv4"]["address"] = interface.sockaddrToStr(interface.addr)
        base_information["ipv4"]["netmask"] = interface.sockaddrToStr(interface.netmask)
        base_information["ipv4"]["gateway"] = "0.0.0.0"
        base_information["ipv4"]["dns"] = _get_dns_info()
        with salt.utils.files.fopen("/proc/net/route", "r") as route_file:
            pattern = re.compile(
                r"^{interface}\t[0]{{8}}\t([0-9A-Z]{{8}})".format(
                    interface=interface.name
                ),
                re.MULTILINE,
            )
            match = pattern.search(route_file.read())
            iface_gateway_hex = None if not match else match.group(1)
        if iface_gateway_hex is not None and len(iface_gateway_hex) == 8:
            base_information["ipv4"]["gateway"] = ".".join(
                [str(int(iface_gateway_hex[i : i + 2], 16)) for i in range(6, -1, -2)]
            )
    return base_information


def _get_interface_info(interface):
    """
    return details about given interface
    """
    adapter_mode = _get_adapter_mode_info(interface.name)
    if adapter_mode == "disabled":
        return _get_base_interface_info(interface)
    elif adapter_mode == "ethercat":
        return _get_ethercat_interface_info(interface)
    return _get_tcpip_interface_info(interface)


def _dict_to_string(dictionary):
    """
    converts a dictionary object into a list of strings
    """
    ret = ""
    for key, val in sorted(dictionary.items()):
        if isinstance(val, dict):
            for line in _dict_to_string(val):
                ret += str(key) + "-" + line + "\n"
        elif isinstance(val, list):
            text = " ".join([str(item) for item in val])
            ret += str(key) + ": " + text + "\n"
        else:
            ret += str(key) + ": " + str(val) + "\n"
    return ret.splitlines()


def _get_info(interface):
    """
    Return information about an interface even if it's not associated with a service.

    :param interface: interface label
    """
    service = _interface_to_service(interface.name)
    if service is not None:
        return _get_service_info(service)
    return _get_static_info(interface)


def get_interfaces_details():
    """
    Get details about all the interfaces on the minion

    :return: information about all interfaces omitting loopback
    :rtype: dictionary

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_interfaces_details
    """
    _interfaces = [
        interface
        for interface in pyiface.getIfaces()
        if interface.flags & IFF_LOOPBACK == 0
    ]
    if __grains__["lsb_distrib_id"] == "nilrt":
        return {"interfaces": list(map(_get_interface_info, _interfaces))}
    return {"interfaces": list(map(_get_info, _interfaces))}


def _change_state_legacy(interface, new_state):
    """
    Enable or disable an interface on a legacy distro

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param interface: interface label
    :param new_state: up or down
    :return: True if the service was enabled, otherwise an exception will be thrown.
    :rtype: bool
    """
    initial_mode = _get_adapter_mode_info(interface)
    _save_config(interface, "Mode", "TCPIP" if new_state == "up" else "Disabled")
    if initial_mode == "ethercat":
        __salt__["system.set_reboot_required_witnessed"]()
    else:
        out = __salt__["cmd.run_all"](f"ip link set {interface} {new_state}")
        if out["retcode"] != 0:
            msg = "Couldn't {} interface {}. Error: {}".format(
                "enable" if new_state == "up" else "disable", interface, out["stderr"]
            )
            raise salt.exceptions.CommandExecutionError(msg)
    return True


def _change_dhcp_config(interface, enable_dhcp=True, filename=INTERFACES_CONFIG):
    """
    Enable or disable dhcp for an interface which isn't a service (in a config file)

    :param interface: interface label
    :param enable_dhcp: True to enable dhcp and False to disable dhcp. Default is True
    :param filename: Config file name. Default is INTERFACES_CONFIG.
    """
    parser = configparser.ConfigParser()
    parser.optionxform = str
    if os.path.exists(filename):
        try:
            with salt.utils.files.fopen(filename, "r") as config_file:
                parser.readfp(config_file)
        except configparser.MissingSectionHeaderError:
            pass
    interface = pyiface.Interface(name=interface)
    hwaddr = interface.hwaddr[:-1]
    hwaddr_section_number = "".join(hwaddr.split(":"))
    if parser.has_section(f"service_{hwaddr_section_number}"):
        parser.remove_section(f"service_{hwaddr_section_number}")
    parser.add_section(f"service_{hwaddr_section_number}")
    parser.set(f"service_{hwaddr_section_number}", "MAC", hwaddr)
    parser.set(
        f"service_{hwaddr_section_number}",
        "Name",
        f"ethernet_cable_{hwaddr_section_number}",
    )
    parser.set(f"service_{hwaddr_section_number}", "Type", "ethernet")
    if enable_dhcp:
        parser.set(f"service_{hwaddr_section_number}", "IPv4.Method", "dhcp")
        parser.set(f"service_{hwaddr_section_number}", "AutoConnect", "true")
        parser.set(f"service_{hwaddr_section_number}", "Nameservers", "''")
    else:
        parser.set(f"service_{hwaddr_section_number}", "IPv4", "off")
    with salt.utils.files.fopen(filename, "w") as config_file:
        parser.write(config_file)
    return True


def _change_state(interface, new_state):
    """
    Enable or disable an interface

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param interface: interface label
    :param new_state: up or down
    :return: True if the service was enabled, otherwise an exception will be thrown.
    :rtype: bool
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        return _change_state_legacy(interface, new_state)
    if interface in [x.name for x in pyiface.getIfaces()]:
        return (
            _change_dhcp_config(interface)
            if new_state == "up"
            else _change_dhcp_config(interface, False)
        )
    raise salt.exceptions.CommandExecutionError(f"Invalid interface name: {interface}")


def up(interface, iface_type=None):  # pylint: disable=invalid-name,unused-argument
    """
    Enable the specified interface

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the service was enabled, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.up interface-label
    """
    return _change_state(interface, "up")


def enable(interface):
    """
    Enable the specified interface

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the service was enabled, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.enable interface-label
    """
    return up(interface)


def down(interface, iface_type=None):  # pylint: disable=unused-argument
    """
    Disable the specified interface

    Change adapter mode to Disabled. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the service was disabled, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.down interface-label
    """
    return _change_state(interface, "down")


def disable(interface):
    """
    Disable the specified interface

    Change adapter mode to Disabled. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the service was disabled, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.disable interface-label
    """
    return down(interface)


def _save_config(section, token, value):
    """
    Helper function to persist a configuration in the ini file
    """
    cmd = NIRTCFG_PATH
    cmd += f" --set section={section},token='{token}',value='{value}'"
    if __salt__["cmd.run_all"](cmd)["retcode"] != 0:
        exc_msg = f"Error: could not set {token} to {value} for {section}\n"
        raise salt.exceptions.CommandExecutionError(exc_msg)


def set_ethercat(interface, master_id):
    """
    Configure specified adapter to use EtherCAT adapter mode. If successful, the target will need reboot if it doesn't
    already use EtherCAT adapter mode, otherwise will return true.

    :param interface: interface label
    :param master_id: EtherCAT Master ID
    :return: True if the settings were applied, otherwise an exception will be thrown.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.set_ethercat interface-label master-id
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        initial_mode = _get_adapter_mode_info(interface)
        _save_config(interface, "Mode", NIRTCFG_ETHERCAT)
        _save_config(interface, "MasterID", master_id)
        if initial_mode != "ethercat":
            __salt__["system.set_reboot_required_witnessed"]()
        return True
    raise salt.exceptions.CommandExecutionError("EtherCAT is not supported")


def _restart(interface):
    """
    Disable and enable an interface
    """
    disable(interface)
    enable(interface)


def set_dhcp_linklocal_all(interface):
    """
    Configure specified adapter to use DHCP with linklocal fallback

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the settings were applied, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.set_dhcp_linklocal_all interface-label
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        initial_mode = _get_adapter_mode_info(interface)
        _save_config(interface, "Mode", "TCPIP")
        _save_config(interface, "dhcpenabled", "1")
        _save_config(interface, "linklocalenabled", "1")
        if initial_mode == "ethercat":
            __salt__["system.set_reboot_required_witnessed"]()
        else:
            _restart(interface)
        return True
    if interface in [x.name for x in pyiface.getIfaces()]:
        return _change_dhcp_config(interface)
    raise salt.exceptions.CommandExecutionError(f"Invalid interface name: {interface}")


def set_dhcp_only_all(interface):
    """
    Configure specified adapter to use DHCP only

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the settings were applied, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.dhcp_only_all interface-label
    """
    if not __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version")
    initial_mode = _get_adapter_mode_info(interface)
    _save_config(interface, "Mode", "TCPIP")
    _save_config(interface, "dhcpenabled", "1")
    _save_config(interface, "linklocalenabled", "0")
    if initial_mode == "ethercat":
        __salt__["system.set_reboot_required_witnessed"]()
    else:
        _restart(interface)
    return True


def set_linklocal_only_all(interface):
    """
    Configure specified adapter to use linklocal only

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :return: True if the settings were applied, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.linklocal_only_all interface-label
    """
    if not __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version")
    initial_mode = _get_adapter_mode_info(interface)
    _save_config(interface, "Mode", "TCPIP")
    _save_config(interface, "dhcpenabled", "0")
    _save_config(interface, "linklocalenabled", "1")
    if initial_mode == "ethercat":
        __salt__["system.set_reboot_required_witnessed"]()
    else:
        _restart(interface)
    return True


def _configure_static_interface(interface, **settings):
    """
    Configure an interface that is not detected as a service by Connman (i.e. link is down)

    :param interface: interface label
    :param settings:
            - ip
            - netmask
            - gateway
            - dns
            - name
    :return: True if settings were applied successfully.
    :rtype: bool
    """
    interface = pyiface.Interface(name=interface)
    parser = configparser.ConfigParser()
    parser.optionxform = str
    if os.path.exists(INTERFACES_CONFIG):
        try:
            with salt.utils.files.fopen(INTERFACES_CONFIG, "r") as config_file:
                parser.readfp(config_file)
        except configparser.MissingSectionHeaderError:
            pass
    hwaddr = interface.hwaddr[:-1]
    hwaddr_section_number = "".join(hwaddr.split(":"))
    if parser.has_section(f"service_{hwaddr_section_number}"):
        parser.remove_section(f"service_{hwaddr_section_number}")
    parser.add_section(f"service_{hwaddr_section_number}")
    ip_address = settings.get("ip", "0.0.0.0")
    netmask = settings.get("netmask", "0.0.0.0")
    gateway = settings.get("gateway", "0.0.0.0")
    dns_servers = settings.get("dns", "''")
    name = settings.get("name", f"ethernet_cable_{hwaddr_section_number}")
    parser.set(
        f"service_{hwaddr_section_number}",
        "IPv4",
        f"{ip_address}/{netmask}/{gateway}",
    )
    parser.set(f"service_{hwaddr_section_number}", "Nameservers", dns_servers)
    parser.set(f"service_{hwaddr_section_number}", "Name", name)
    parser.set(f"service_{hwaddr_section_number}", "MAC", hwaddr)
    parser.set(f"service_{hwaddr_section_number}", "Type", "ethernet")
    parser.set(f"service_{hwaddr_section_number}", "IPv4.method", "manual")
    with salt.utils.files.fopen(INTERFACES_CONFIG, "w") as config_file:
        parser.write(config_file)
    return True


def set_static_all(interface, address, netmask, gateway, nameservers=None):
    """
    Configure specified adapter to use ipv4 manual settings

    Change adapter mode to TCP/IP. If previous adapter mode was EtherCAT, the target will need reboot.

    :param str interface: interface label
    :param str address: ipv4 address
    :param str netmask: ipv4 netmask
    :param str gateway: ipv4 gateway
    :param str nameservers: list of nameservers servers separated by spaces (Optional)
    :return: True if the settings were applied, otherwise an exception will be thrown.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' ip.set_static_all interface-label address netmask gateway nameservers
    """
    validate, msg = _validate_ipv4([address, netmask, gateway])
    if not validate:
        raise salt.exceptions.CommandExecutionError(msg)
    if nameservers:
        validate, msg = _space_delimited_list(nameservers)
        if not validate:
            raise salt.exceptions.CommandExecutionError(msg)
        if not isinstance(nameservers, list):
            nameservers = nameservers.split(" ")
    if __grains__["lsb_distrib_id"] == "nilrt":
        initial_mode = _get_adapter_mode_info(interface)
        _save_config(interface, "Mode", "TCPIP")
        _save_config(interface, "dhcpenabled", "0")
        _save_config(interface, "linklocalenabled", "0")
        _save_config(interface, "IP_Address", address)
        _save_config(interface, "Subnet_Mask", netmask)
        _save_config(interface, "Gateway", gateway)
        if nameservers:
            _save_config(interface, "DNS_Address", nameservers[0])
        if initial_mode == "ethercat":
            __salt__["system.set_reboot_required_witnessed"]()
        else:
            _restart(interface)
        return True

    service = _interface_to_service(interface)
    if not service:
        if interface in [x.name for x in pyiface.getIfaces()]:
            return _configure_static_interface(
                interface,
                **{
                    "ip": address,
                    "dns": ",".join(nameservers) if nameservers else "",
                    "netmask": netmask,
                    "gateway": gateway,
                },
            )
        raise salt.exceptions.CommandExecutionError(
            f"Invalid interface name: {interface}"
        )
    service = pyconnman.ConnService(os.path.join(SERVICE_PATH, service))
    ipv4 = service.get_property("IPv4.Configuration")
    ipv4["Method"] = dbus.String("manual", variant_level=1)
    ipv4["Address"] = dbus.String(f"{address}", variant_level=1)
    ipv4["Netmask"] = dbus.String(f"{netmask}", variant_level=1)
    ipv4["Gateway"] = dbus.String(f"{gateway}", variant_level=1)
    try:
        service.set_property("IPv4.Configuration", ipv4)
        if nameservers:
            service.set_property(
                "Nameservers.Configuration",
                [dbus.String(f"{d}") for d in nameservers],
            )
    except Exception as exc:  # pylint: disable=broad-except
        exc_msg = "Couldn't set manual settings for service: {}\nError: {}\n".format(
            service, exc
        )
        raise salt.exceptions.CommandExecutionError(exc_msg)
    return True


def get_interface(iface):
    """
    Returns details about given interface.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_interface eth0
    """
    _interfaces = get_interfaces_details()
    for _interface in _interfaces["interfaces"]:
        if _interface["connectionid"] == iface:
            return _dict_to_string(_interface)
    return None


def build_interface(iface, iface_type, enabled, **settings):
    """
    Build an interface script for a network interface.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_interface eth0 eth <settings>
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version.")
    if iface_type != "eth":
        raise salt.exceptions.CommandExecutionError(
            f"Interface type not supported: {iface_type}:"
        )

    if (
        "proto" not in settings or settings["proto"] == "dhcp"
    ):  # default protocol type used is dhcp
        set_dhcp_linklocal_all(iface)
    elif settings["proto"] != "static":
        exc_msg = "Protocol type: {} is not supported".format(settings["proto"])
        raise salt.exceptions.CommandExecutionError(exc_msg)
    else:
        address = settings["ipaddr"]
        netmask = settings["netmask"]
        gateway = settings["gateway"]
        dns = []
        for key, val in settings.items():
            if "dns" in key or "domain" in key:
                dns += val
        set_static_all(iface, address, netmask, gateway, dns)

    if enabled:
        up(iface)

    return get_interface(iface)


def build_network_settings(**settings):
    """
    Build the global network script.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_network_settings <settings>
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version.")
    changes = []
    if "networking" in settings:
        if settings["networking"] in _CONFIG_TRUE:
            __salt__["service.enable"]("connman")
        else:
            __salt__["service.disable"]("connman")

    if "hostname" in settings:
        new_hostname = settings["hostname"].split(".", 1)[0]
        settings["hostname"] = new_hostname
        old_hostname = __salt__["network.get_hostname"]
        if new_hostname != old_hostname:
            __salt__["network.mod_hostname"](new_hostname)
            changes.append(f"hostname={new_hostname}")

    return changes


def get_network_settings():
    """
    Return the contents of the global network script.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_network_settings
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version.")
    settings = []
    networking = "no" if _get_state() == "offline" else "yes"
    settings.append(f"networking={networking}")
    hostname = __salt__["network.get_hostname"]
    settings.append(f"hostname={hostname}")
    return settings


def apply_network_settings(**settings):
    """
    Apply global network configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.apply_network_settings
    """
    if __grains__["lsb_distrib_id"] == "nilrt":
        raise salt.exceptions.CommandExecutionError("Not supported in this version.")
    if "require_reboot" not in settings:
        settings["require_reboot"] = False

    if "apply_hostname" not in settings:
        settings["apply_hostname"] = False

    hostname_res = True
    if settings["apply_hostname"] in _CONFIG_TRUE:
        if "hostname" in settings:
            hostname_res = __salt__["network.mod_hostname"](settings["hostname"])
        else:
            log.warning(
                "The network state sls is trying to apply hostname "
                "changes but no hostname is defined."
            )
            hostname_res = False

    res = True
    if settings["require_reboot"] in _CONFIG_TRUE:
        log.warning(
            "The network state sls is requiring a reboot of the system to "
            "properly apply network configuration."
        )
        res = True
    else:
        stop = __salt__["service.stop"]("connman")
        time.sleep(2)
        res = stop and __salt__["service.start"]("connman")

    return hostname_res and res



"""
Connection module for Amazon SQS

.. versionadded:: 2014.7.0

:configuration: This module accepts explicit sqs credentials but can also utilize
    IAM roles assigned to the instance through Instance Profiles. Dynamic
    credentials are then automatically obtained from AWS API and no further
    configuration is necessary. More information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        sqs.keyid: GKTADJGHEIQSXMKKRBJ08H
        sqs.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        sqs.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto3
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging
import urllib.parse

import salt.utils.json
import salt.utils.versions

log = logging.getLogger(__name__)

__func_alias__ = {
    "list_": "list",
}

try:
    # pylint: disable=unused-import
    import boto3
    import botocore

    # pylint: enable=unused-import
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False


def __virtual__():
    """
    Only load if boto3 libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs()
    if has_boto_reqs is True:
        __utils__["boto3.assign_funcs"](__name__, "sqs")
    return has_boto_reqs


def _preprocess_attributes(attributes):
    """
    Pre-process incoming queue attributes before setting them
    """
    if isinstance(attributes, str):
        attributes = salt.utils.json.loads(attributes)

    def stringified(val):
        # Some attributes take full json policy documents, but they take them
        # as json strings. Convert the value back into a json string.
        if isinstance(val, dict):
            return salt.utils.json.dumps(val)
        return val

    return {attr: stringified(val) for attr, val in attributes.items()}


def exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if a queue exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.exists myqueue region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        conn.get_queue_url(QueueName=name)
    except botocore.exceptions.ClientError as e:
        missing_code = "AWS.SimpleQueueService.NonExistentQueue"
        if e.response.get("Error", {}).get("Code") == missing_code:
            return {"result": False}
        return {"error": __utils__["boto3.get_error"](e)}
    return {"result": True}


def create(
    name,
    attributes=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an SQS queue.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.create myqueue region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if attributes is None:
        attributes = {}
    attributes = _preprocess_attributes(attributes)

    try:
        conn.create_queue(QueueName=name, Attributes=attributes)
    except botocore.exceptions.ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}
    return {"result": True}


def delete(name, region=None, key=None, keyid=None, profile=None):
    """
    Delete an SQS queue.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.delete myqueue region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        url = conn.get_queue_url(QueueName=name)["QueueUrl"]
        conn.delete_queue(QueueUrl=url)
    except botocore.exceptions.ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}
    return {"result": True}


def list_(prefix="", region=None, key=None, keyid=None, profile=None):
    """
    Return a list of the names of all visible queues.

    .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.list region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    def extract_name(queue_url):
        # Note: this logic taken from boto, so should be safe
        return urllib.parse.urlparse(queue_url).path.split("/")[2]

    try:
        r = conn.list_queues(QueueNamePrefix=prefix)
        # The 'QueueUrls' attribute is missing if there are no queues
        urls = r.get("QueueUrls", [])
        return {"result": [extract_name(url) for url in urls]}
    except botocore.exceptions.ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def get_attributes(name, region=None, key=None, keyid=None, profile=None):
    """
    Return attributes currently set on an SQS queue.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.get_attributes myqueue
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        url = conn.get_queue_url(QueueName=name)["QueueUrl"]
        r = conn.get_queue_attributes(QueueUrl=url, AttributeNames=["All"])
        return {"result": r["Attributes"]}
    except botocore.exceptions.ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}


def set_attributes(
    name,
    attributes,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Set attributes on an SQS queue.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_sqs.set_attributes myqueue '{ReceiveMessageWaitTimeSeconds: 20}' region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    attributes = _preprocess_attributes(attributes)

    try:
        url = conn.get_queue_url(QueueName=name)["QueueUrl"]
        conn.set_queue_attributes(QueueUrl=url, Attributes=attributes)
    except botocore.exceptions.ClientError as e:
        return {"error": __utils__["boto3.get_error"](e)}
    return {"result": True}



import pytest

import salt.modules.win_appx as win_appx
from tests.support.mock import MagicMock, call, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {win_appx: {}}


def test__pkg_list_empty():
    assert win_appx._pkg_list("") is None


def test__pkg_list_single():
    raw = {
        "Name": "MicrosoftTeams",
        "Version": "22042.702.1226.2352",
        "PackageFullName": "MicrosoftTeams_22042.702.1226.2352_x64__8wekyb3d8bbwe",
        "PackageFamilyName": "MicrosoftTeams_8wekyb3d8bbwe",
    }
    assert win_appx._pkg_list(raw=raw) == ["MicrosoftTeams"]


def test__pkg_list_multiple():
    raw = [
        {
            "Name": "MicrosoftTeams",
            "Version": "22042.702.1226.2352",
            "PackageFullName": "MicrosoftTeams_22042.702.1226.2352_x64__8wekyb3d8bbwe",
            "PackageFamilyName": "MicrosoftTeams_8wekyb3d8bbwe",
        },
        {
            "Name": "Microsoft.BingWeather",
            "Version": "4.53.51361.0",
            "PackageFullName": "Microsoft.BingWeather_4.53.51361.0_x64__8wekyb3d8bbwe",
            "PackageFamilyName": "Microsoft.BingWeather_8wekyb3d8bbwe",
        },
    ]
    assert win_appx._pkg_list(raw=raw) == ["MicrosoftTeams", "Microsoft.BingWeather"]


def test__pkg_list_single_field():
    raw = {
        "Name": "MicrosoftTeams",
        "Version": "22042.702.1226.2352",
        "PackageFullName": "MicrosoftTeams_22042.702.1226.2352_x64__8wekyb3d8bbwe",
        "PackageFamilyName": "MicrosoftTeams_8wekyb3d8bbwe",
    }
    assert win_appx._pkg_list(raw=raw, field="PackageFamilyName") == [
        "MicrosoftTeams_8wekyb3d8bbwe"
    ]


def test__pkg_list_multiple_field():
    raw = [
        {
            "Name": "MicrosoftTeams",
            "Version": "22042.702.1226.2352",
            "PackageFullName": "MicrosoftTeams_22042.702.1226.2352_x64__8wekyb3d8bbwe",
            "PackageFamilyName": "MicrosoftTeams_8wekyb3d8bbwe",
        },
        {
            "Name": "Microsoft.BingWeather",
            "Version": "4.53.51361.0",
            "PackageFullName": "Microsoft.BingWeather_4.53.51361.0_x64__8wekyb3d8bbwe",
            "PackageFamilyName": "Microsoft.BingWeather_8wekyb3d8bbwe",
        },
    ]
    assert win_appx._pkg_list(raw=raw, field="PackageFamilyName") == [
        "MicrosoftTeams_8wekyb3d8bbwe",
        "Microsoft.BingWeather_8wekyb3d8bbwe",
    ]


def test_list():
    mock_run_dict = MagicMock()
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict):
        win_appx.list_("*test*")
    cmd = [
        "Get-AppxPackage -AllUsers -PackageTypeFilter Bundle -Name *test*",
        'Where-Object {$_.name -notlike "Microsoft.WindowsStore*"}',
        "Where-Object -Property IsFramework -eq $false",
        "Where-Object -Property NonRemovable -eq $false",
        "Sort-Object Name",
    ]
    mock_run_dict.assert_called_once_with(" | ".join(cmd))


def test_list_field_none():
    mock_run_dict = MagicMock()
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict):
        win_appx.list_("*test*", field=None)
    cmd = [
        "Get-AppxPackage -AllUsers -PackageTypeFilter Bundle -Name *test*",
        'Where-Object {$_.name -notlike "Microsoft.WindowsStore*"}',
        "Where-Object -Property IsFramework -eq $false",
        "Where-Object -Property NonRemovable -eq $false",
        "Sort-Object Name",
        "Select Name, Version, PackageFullName, PackageFamilyName, IsBundle, IsFramework",
    ]
    mock_run_dict.assert_called_once_with(" | ".join(cmd))


def test_list_other_options_flipped():
    mock_run_dict = MagicMock()
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict):
        win_appx.list_("*test*", include_store=True, frameworks=True, bundles=False)
    cmd = [
        "Get-AppxPackage -AllUsers -Name *test*",
        "Where-Object -Property NonRemovable -eq $false",
        "Sort-Object Name",
    ]
    mock_run_dict.assert_called_once_with(" | ".join(cmd))


def test_remove():
    mock_run_dict = MagicMock()
    mock_list_return = {
        "Name": "Microsoft.BingWeather",
        "PackageFullName": "Microsoft.BingWeather_full_name",
        "IsBundle": True,
    }
    mock_list = MagicMock(return_value=mock_list_return)
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*test*") is True
    cmd = "Remove-AppxPackage -AllUsers -Package Microsoft.BingWeather_full_name"
    mock_run_dict.assert_called_with(cmd)


def test_remove_duplicate():
    mock_run_dict = MagicMock()
    mock_list_return_1 = [
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_full_name_1",
            "IsBundle": False,
        },
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_full_name_2",
            "IsBundle": False,
        },
    ]
    mock_list_return_2 = [
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_full_name_1",
            "IsBundle": True,
        },
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_full_name_2",
            "IsBundle": True,
        },
    ]
    mock_list_return_3 = [
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_full_name_2",
            "IsBundle": True,
        },
    ]
    mock_list = MagicMock(
        side_effect=[mock_list_return_1, mock_list_return_2, mock_list_return_3]
    )
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*bingweather*") is True
    cmd_1 = "Remove-AppxPackage -AllUsers -Package Microsoft.BingWeather_full_name_1"
    cmd_2 = "Remove-AppxPackage -AllUsers -Package Microsoft.BingWeather_full_name_2"
    mock_run_dict.assert_has_calls([call(cmd_1), call(cmd_2)])


def test_remove_deprovision_only():
    mock_run_dict = MagicMock()
    mock_list_return = {
        "Name": "Microsoft.BingWeather",
        "PackageFullName": "Microsoft.BingWeather_full_name",
        "IsBundle": True,
    }
    mock_list = MagicMock(return_value=mock_list_return)
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*test*", deprovision_only=True) is True
    cmd = "Remove-AppxProvisionedPackage -Online -PackageName Microsoft.BingWeather_full_name"
    mock_run_dict.assert_called_with(cmd)


def test_remove_non_bundle():
    mock_run_dict = MagicMock()
    mock_list_return = [
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_non_bundle",
            "IsBundle": False,
        },
        {
            "Name": "Microsoft.BingWeather",
            "PackageFullName": "Microsoft.BingWeather_bundle",
            "IsBundle": True,
        },
    ]
    mock_list = MagicMock(side_effect=mock_list_return)
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*test*", deprovision_only=True) is True
    cmd = "Remove-AppxProvisionedPackage -Online -PackageName Microsoft.BingWeather_bundle"
    mock_run_dict.assert_called_with(cmd)


def test_remove_not_found_empty_dict():
    mock_run_dict = MagicMock()
    mock_list_return = {}
    mock_list = MagicMock(return_value=mock_list_return)
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*test*", deprovision_only=True) is None


def test_remove_not_found_none():
    mock_run_dict = MagicMock()
    mock_list_return = None
    mock_list = MagicMock(return_value=mock_list_return)
    with patch("salt.utils.win_pwsh.run_dict", mock_run_dict), patch.object(
        win_appx, "list_", mock_list
    ):
        assert win_appx.remove("*test*", deprovision_only=True) is None


def test_list_deprovisioned():
    mock_list_keys = MagicMock(return_value=["Deprovisioned1", "Deprovisioned2"])
    with patch("salt.utils.win_reg.list_keys", mock_list_keys):
        expected = ["Deprovisioned1", "Deprovisioned2"]
        assert win_appx.list_deprovisioned() == expected


def test_list_deprovisioned_query():
    mock_list_keys = MagicMock(return_value=["Deprovisioned1", "Deprovisioned2"])
    with patch("salt.utils.win_reg.list_keys", mock_list_keys):
        expected = ["Deprovisioned1"]
        assert win_appx.list_deprovisioned(query="*ed1*") == expected


def test_install():
    mock_dism = MagicMock(return_value={"retcode": 0})
    with patch.dict(win_appx.__salt__, {"dism.add_provisioned_package": mock_dism}):
        assert win_appx.install("C:\\Test.appx") is True
    mock_dism.assert_called_once_with("C:\\Test.appx")
